"""
Structure for efficient access and modification of bounding boxes with
associated scores and class labels. Builds on top of the `kwimage.Boxes`
structure.

Also can optionally incorporate `kwimage.PolygonList` for segmentation masks
and `kwimage.PointsList` for keypoints.


If you want to visualize boxes and scores you can do this:
    >>> # Given data
    >>> data = np.random.rand(10, 4) * 224
    >>> scores = np.random.rand(10,)
    >>> class_idxs = np.random.randint(0, 3, size=10)
    >>> classes = ['class1', 'class2', 'class3']
    >>> #
    >>> # Wrap your data with a Detections object
    >>> import kwimage
    >>> dets = kwimage.Detections(
    >>>     boxes=kwimage.Boxes(data, format='xywh'),
    >>>     scores=scores,
    >>>     class_idxs=class_idxs,
    >>>     classes=classes,
    >>> )
    >>> # xdoctest: +REQUIRES(module:kwplot)
    >>> dets.draw()
    >>> import matplotlib.pyplot as plt
    >>> plt.gca().set_xlim(0, 224)
    >>> plt.gca().set_ylim(0, 224)

"""
import numpy as np
import ubelt as ub
import sys
from kwimage.structs import boxes as _boxes
from kwimage.structs import _generic

# try:
#     import torch
# except Exception:
#     torch = None


class _DetDrawMixin:
    """
    Non critical methods for visualizing detections
    """
    def draw(self, color='blue', alpha=None, labels=True, centers=False, lw=2,
             fill=False, ax=None, radius=5, kpts=True, sseg=True,
             setlim=False, boxes=True):
        """
        Draws boxes using matplotlib

        Example:
            >>> # xdoctest: +REQUIRES(module:kwplot)
            >>> self = Detections.random(num=10, scale=512.0, rng=0, classes=['a', 'b', 'c'])
            >>> self.boxes.translate((-128, -128), inplace=True)
            >>> image = (np.random.rand(256, 256) * 255).astype(np.uint8)
            >>> # xdoctest: +REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.autompl()
            >>> fig = kwplot.figure(fnum=1, doclf=True)
            >>> kwplot.imshow(image)
            >>> # xdoctest: +REQUIRES(--show)
            >>> self.draw(color='blue', alpha=None)
            >>> # xdoctest: +REQUIRES(--show)
            >>> for o in fig.findobj():  # http://matplotlib.1069221.n5.nabble.com/How-to-turn-off-all-clipping-td1813.html
            >>>     o.set_clip_on(False)
            >>> kwplot.show_if_requested()
        """
        segmentations = self.data.get('segmentations', None)
        if sseg and segmentations is not None:
            segmentations.draw(color=color, alpha=.4)

        labels = self._make_labels(labels)
        alpha = self._make_alpha(alpha)
        if boxes:
            self.boxes.draw(labels=labels, color=color, alpha=alpha, fill=fill,
                            centers=centers, ax=ax, lw=lw)

        keypoints = self.data.get('keypoints', None)
        if kpts and keypoints is not None:
            keypoints.draw(color=color, radius=radius)

        if setlim:
            x1, y1, x2, y2 = self.boxes.to_ltrb().components
            xmax = x2.max()
            xmin = x1.min()
            ymax = y2.max()
            ymin = y1.min()
            import matplotlib.pyplot as plt
            ax = plt.gca()
            ax.set_xlim(xmin, xmax)
            ax.set_ylim(ymin, ymax)

    def draw_on(self, image=None, color='blue', alpha=None, labels=True, radius=5,
                kpts=True, sseg=True, boxes=True, ssegkw=None,
                label_loc='top_left', thickness=2):
        """
        Draws boxes directly on the image using OpenCV

        Args:
            image (ndarray[Any, UInt8]): must be in uint8 format

            color (str | Any | List[Any]):
                one color for all boxes or a list of colors for each box.
                Or the string "classes", in which case it will use a
                different color for each class (specified in the classes object
                if possible).  Extended types: str | ColorLike |
                List[ColorLike]

            alpha (float): Transparency of overlay. can be a scalar or a list
                for each box

            labels (bool | str | List[str]):
                if True, use categorie names as the labels. See _make_labels
                for details. Otherwise a manually specified text label for each
                box.

            boxes (bool): if True draw the boxes

            kpts (bool): if True draw the keypoints

            sseg (bool): if True draw the segmentations

            ssegkw (dict): extra arguments passed to `segmentations.draw_on`

            radius (float): passed to `keypoints.draw_on`

            label_loc (str): indicates where labels (if specified) should be
                drawn. passed to `boxes.draw_on`

            thickness (int): rectangle thickness, negative values
                will draw a filled rectangle. passed to `boxes.draw_on`.
                Defaults to 2.

        Returns:
            ndarray[Any, UInt8]: image with labeled boxes drawn on it

        CommandLine:
            xdoctest -m kwimage.structs.detections _DetDrawMixin.draw_on:1 --profile --show

        Example:
            >>> # xdoctest: +REQUIRES(module:kwplot)
            >>> import kwimage
            >>> import kwplot
            >>> self = kwimage.Detections.random(num=10, scale=512, rng=0)
            >>> image = (np.random.rand(512, 512) * 255).astype(np.uint8)
            >>> image2 = self.draw_on(image, color='blue')
            >>> # xdoctest: +REQUIRES(--show)
            >>> kwplot.autompl()
            >>> kwplot.figure(fnum=2000, doclf=True)
            >>> kwplot.imshow(image2, title='test basic detection drawing')
            >>> kwplot.show_if_requested()

        Example:
            >>> # xdoctest: +REQUIRES(module:kwplot)
            >>> import kwimage
            >>> from kwimage.structs.detections import *  # NOQA
            >>> import kwplot
            >>> self = kwimage.Detections.random(num=10, scale=512, rng=0)
            >>> image = (np.random.rand(512, 512) * 255).astype(np.uint8)
            >>> image2 = self.draw_on(image, color='classes')
            >>> # xdoctest: +REQUIRES(--show)
            >>> kwplot.autompl()
            >>> kwplot.figure(fnum=2000, doclf=True)
            >>> kwplot.imshow(image2, title='test color by classes')
            >>> kwplot.show_if_requested()

        Example:
            >>> # xdoctest: +REQUIRES(module:kwplot)
            >>> # xdoctest: +REQUIRES(--profile)
            >>> import kwimage
            >>> import kwplot
            >>> self = kwimage.Detections.random(num=100, scale=512, rng=0, keypoints=True, segmentations=True)
            >>> image = (np.random.rand(512, 512) * 255).astype(np.uint8)
            >>> image2 = self.draw_on(image, color='blue')
            >>> # xdoctest: +REQUIRES(--show)
            >>> kwplot.figure(fnum=2000, doclf=True)
            >>> kwplot.autompl()
            >>> kwplot.imshow(image2, title='test keypoints and segmentations')
            >>> kwplot.show_if_requested()

        Example:
            >>> # xdoctest: +REQUIRES(module:kwplot)
            >>> # Test that boxes with tiny scores are drawn correctly
            >>> import kwimage
            >>> import kwplot
            >>> self = kwimage.Detections.random(num=10, scale=512, rng=0)
            >>> self.data['scores'][:] = 1.23e-8
            >>> image = (np.random.rand(512, 512) * 255).astype(np.uint8)
            >>> image2 = self.draw_on(image, color='blue')
            >>> # xdoctest: +REQUIRES(--show)
            >>> kwplot.autompl()
            >>> kwplot.figure(fnum=2000, doclf=True, title='test tiny scores')
            >>> kwplot.imshow(image2)
            >>> kwplot.show_if_requested()

        Example:
            >>> # xdoctest: +REQUIRES(module:kwplot)
            >>> # Test that alpha can be passed as an iterable.
            >>> import kwimage
            >>> import kwplot
            >>> self = kwimage.Detections.random(num=3, scale=512, rng=0)
            >>> canvas = self.draw_on(None, color='blue', alpha=[0, 0.5, 1.0])
            >>> unique_values = np.unique(canvas)
            >>> assert len(unique_values) == 3, ub.paragraph(
                '''2 non-zero alpha colors plus background should have 3 unique
                colors''')
            >>> # xdoctest: +REQUIRES(--show)
            >>> kwplot.autompl()
            >>> kwplot.imshow(canvas, fnum=2000, doclf=True, title='test custom alphas')
            >>> kwplot.show_if_requested()

        Ignore:
            # xdoctest: +REQUIRES(module:kwplot)
            # xdoctest: +REQUIRES(--profile)
            import kwplot

            import kwimage
            self = kwimage.Detections.random(num=2, scale=512, rng=0, keypoints=True, segmentations=True)
            image = (np.random.rand(512, 512)).astype(np.float32)
            self = self.scale(2e6)
            image2 = self.draw_on(image, color='blue')

            >>> # xdoctest: +REQUIRES(--show)
            >>> kwplot.figure(fnum=2000, doclf=True)
            >>> kwplot.autompl()
            >>> kwplot.imshow(image2)
            >>> kwplot.show_if_requested()

        Ignore:
            import xdev
            globals().update(xdev.get_func_kwargs(kwimage.Detections.draw_on))
        """
        if image is None:
            # If image is not given, use the boxes to allocate enough
            # room to draw
            bounds = self.boxes.scale(1.5).bounding_box().quantize()
            w = bounds.width.item()
            h = bounds.height.item()
            w = h = max(w, h)
            image = np.zeros((h, w, 3), dtype=np.float32)

        labels = self._make_labels(labels)
        alpha = self._make_alpha(alpha)
        color = self._make_colors(color)

        dtype_fixer = _generic._consistent_dtype_fixer(image)

        if len(self):

            segmentations = self.data.get('segmentations', None)
            if sseg and segmentations is not None:
                if ssegkw is None:
                    ssegkw = {
                        'alpha': 0.4,
                        'color': color,
                    }
                image = segmentations.draw_on(image, **ssegkw)

            if boxes:
                image = self.boxes.draw_on(image, color=color, alpha=alpha,
                                           labels=labels, label_loc=label_loc,
                                           thickness=thickness)

            keypoints = self.data.get('keypoints', None)
            if kpts and keypoints is not None:
                # image = kwimage.ensure_float01(image)
                image = keypoints.draw_on(image, radius=radius, color=color)
                # kwimage.ensure_float01(image)

        image = dtype_fixer(image, copy=False)
        return image

    def _make_colors(self, color):
        """
        Handles special settings of color.

        If color == 'classes', then choose a distinct color for each category
        """
        # Draw each category as a different color
        if color == 'classes':
            import kwimage
            backup_color = 'blue'
            class_idxs = self.class_idxs
            if class_idxs is None:
                color = backup_color
            else:
                classes = self.classes
                if classes is None:
                    classes = list(range(max(class_idxs) + 1))

                # TODO: allow specified color scheme
                backup_colors = iter(kwimage.Color.distinct(len(classes)))

                # Respect colors stored in classes if given
                if hasattr(classes, 'idx_to_node'):
                    cname_to_color = {
                        cid: cat.get('color', None)
                        for cid, cat in classes.cats.items()
                    }
                    cidx_to_color = [
                        cname_to_color[cname]
                        for cname in classes.idx_to_node
                    ]
                else:
                    cidx_to_color = [None] * len(classes)

                for cidx, c in enumerate(cidx_to_color):
                    if c is None:
                        cidx_to_color[cidx] = next(backup_colors)
                    else:
                        cidx_to_color[cidx] = c
                        # kwimage.Color(c).as01()

                color = [backup_color if cidx is None else cidx_to_color[cidx]
                         for cidx in class_idxs]
        return color

    def _make_alpha(self, alpha):
        """
        Either passes through user specified alpha or chooses a sensible
        default
        """
        if isinstance(alpha, str) and alpha in ['score', 'scores']:
            alpha = np.sqrt(self.scores)
        elif ub.iterable(alpha):
            assert len(alpha) == self.num_boxes()
        else:
            if alpha is None or alpha is False:
                alpha = 1.0
            alpha = [float(alpha)] * self.num_boxes()
        return alpha

    def _make_labels(self, labels):
        """
        Either passes through user specified labels or chooses a sensible
        default
        """
        def _fixsore(s):
            return float('nan') if s is None else s

        if labels:
            if labels is True or isinstance(labels, int) and labels:
                parts = []
                if self.data.get('class_idxs', None) is not None:
                    parts.append('class')
                elif self.data.get('cids', None) is not None:
                    parts.append('class')
                # Choose sensible default
                if self.data.get('scores', None) is not None:
                    parts.append('score')
                labels = '+'.join(parts)

            if isinstance(labels, str):
                if labels in ['class', 'class+score']:
                    if 'class_idxs' in self.data:
                        if self.classes:
                            identifers = ['None' if cx is None else self.classes[cx] for cx in self.class_idxs]
                        else:
                            identifers = ['cx={}'.format(cx) for cx in self.class_idxs]
                    elif 'cids' in self.data:
                        if self.classes and hasattr(self.classes, 'id_to_node'):
                            identifers = ['None' if cid is None else self.classes.id_to_node[cid] for cid in self.data['cids']]
                        else:
                            identifers = ['cid={}'.format(cid) for cid in self.data['cids']]
                    else:
                        # Cant determine label for class
                        identifers = ['?' for _ in range(len(self))]

                if labels in ['class']:
                    labels = identifers
                elif labels in ['score']:
                    labels = ['{:.4f}'.format(_fixsore(score)) for score in self.scores]
                elif labels in ['class+score']:
                    labels = ['{} @ {:.4f}'.format(cid, _fixsore(score))
                              for cid, score in zip(identifers, self.scores)]
                else:
                    raise KeyError('unknown labels key {!r}'.format(labels))
        return labels


class _DetAlgoMixin:
    """
    Non critical methods for algorithmic manipulation of detections
    """

    def non_max_supression(self, thresh=0.0, perclass=False, impl='auto',
                           daq=False, device_id=None):
        """
        Find high scoring minimally overlapping detections

        Args:
            thresh (float): iou threshold between 0 and 1. A box is removed if
                it overlaps with a previously chosen box by more than this
                threshold. Higher values are are more permissive (more boxes
                are returned). A value of 0 means that returned boxes will have
                no overlap.

            perclass (bool): if True, works on a per-class basis

            impl (str): implementation can be "auto", "python", "cython_cpu",
                "gpu", "torch", or "torchvision". Not all backends may be
                available, see :func:`kwimage.algo.algo_nms.available_nms_impls`
                for what is supported on your system.

            daq (bool | Dict): if False, uses reqgular nms, otherwise uses
                divide and conquer algorithm. If `daq` is a Dict, then
                it is used as the kwargs to `kwimage.daq_spatial_nms`

            device_id : try not to use. only used if impl is gpu

        Returns:
            ndarray[Shape['*'], Integer]: indices of boxes to keep

        SeeAlso:
            :func:`kwimage.algo.algo_nms.non_max_supression`.

        Example:
            >>> import kwimage
            >>> dets1 = kwimage.Detections.random(rng=0).scale((512, 512))
            >>> keep = dets1.non_max_supression(thresh=0.2)
            >>> dets2 = dets1.take(keep)
            >>> # xdoctest: +REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.autompl()
            >>> canvas = np.zeros((512, 512, 3))
            >>> canvas1 = dets1.draw_on(canvas.copy())
            >>> canvas2 = dets2.draw_on(canvas.copy())
            >>> kwplot.figure(fnum=1, pnum=(1, 2, 1))
            >>> kwplot.imshow(canvas1)
            >>> kwplot.figure(fnum=1, pnum=(1, 2, 2))
            >>> kwplot.imshow(canvas2)
        """
        import kwimage
        classes = self.class_idxs if perclass else None

        if len(self) <= 0:
            return []

        ltrb = self.boxes.to_ltrb().data
        scores = self.data.get('scores', None)
        if scores is None:
            scores = np.ones(len(self), dtype=np.float32)
        if daq:
            daqkw = {} if daq is True else daq.copy()
            daqkw['impl'] = daqkw.get('impl', impl)
            daqkw['stop_size'] = daqkw.get('stop_size', 2048)
            daqkw['max_depth'] = daqkw.get('max_depth', 12)
            daqkw['thresh'] = daqkw.get('thresh', thresh)
            if 'diameter' not in daqkw:
                if len(self.boxes) > 0:
                    daqkw['diameter'] = max(self.boxes.width.max(),
                                            self.boxes.height.max())
                else:
                    daqkw['diameter'] = 10  # hack

            keep = kwimage.daq_spatial_nms(ltrb, scores, device_id=device_id,
                                           **daqkw)
        else:
            keep = kwimage.non_max_supression(ltrb, scores, thresh=thresh,
                                              classes=classes, impl=impl,
                                              device_id=device_id)
        return keep

    def non_max_supress(self, thresh=0.0, perclass=False, impl='auto',
                        daq=False):
        """
        Convinience method. Like `non_max_supression`, but returns to supressed
        boxes instead of the indices to keep.
        """
        keep = self.non_max_supression(thresh=thresh, perclass=perclass,
                                       impl=impl, daq=daq)
        return self.take(keep)

    def rasterize(self, bg_size, input_dims, soften=1, tf_data_to_img=None,
                  img_dims=None, exclude=[]):
        """
        Ambiguous conversion from a Heatmap to a Detections object.

        SeeAlso:
            Heatmap.detect

        Returns:
            kwimage.Heatmap: raster-space detections.

        Example:
            >>> # xdoctest: +REQUIRES(module:ndsampler)
            >>> from kwimage.structs.detections import *  # NOQA
            >>> self, iminfo, sampler = Detections.demo()
            >>> image = iminfo['imdata'][:]
            >>> input_dims = iminfo['imdata'].shape[0:2]
            >>> bg_size = [100, 100]
            >>> heatmap = self.rasterize(bg_size, input_dims)
            >>> # xdoctest: +REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.autompl()
            >>> kwplot.figure(fnum=1, pnum=(2, 2, 1))
            >>> heatmap.draw(invert=True)
            >>> kwplot.figure(fnum=1, pnum=(2, 2, 2))
            >>> kwplot.imshow(heatmap.draw_on(image))
            >>> kwplot.figure(fnum=1, pnum=(2, 1, 2))
            >>> kwplot.imshow(heatmap.draw_stacked())
        """
        import kwarray
        import skimage
        import kwimage
        classes = self.meta['classes']

        try:
            bg_idx = classes.index('background')
        except Exception:
            # TODO: might not be right to except this
            bg_idx = 0

        fcn_target = _dets_to_fcmaps(
            self, bg_size=bg_size, input_dims=input_dims, bg_idx=bg_idx,
            soft=False, exclude=exclude)

        if tf_data_to_img is None:
            tf_data_to_img = skimage.transform.AffineTransform(
                scale=(1, 1), translation=(0, 0),
            )

        if img_dims is None:
            img_dims = np.array(input_dims)
        # print(fcn_target.keys())
        # print('fcn_target: ' + ub.urepr(ub.map_vals(lambda x: x.shape, fcn_target), nl=1))

        impl = kwarray.ArrayAPI.coerce(fcn_target['cidx'])

        # class_probs = nh.criterions.focal.one_hot_embedding(
        #     fcn_target['cidx'].reshape(-1),
        #     num_classes=len(classes), dim=1)

        class_idx = fcn_target['cidx']

        if 'class_probs' not in exclude:
            class_probs = kwarray.one_hot_embedding(
                class_idx, num_classes=len(classes), dim=0)

            if soften > 0:
                k = 31
                sigma = 0.3 * ((k - 1) * 0.5 - 1) + 0.8  # opencv formula
                data = impl.contiguous(class_probs.T)
                import cv2
                if data.size > 0:
                    cv2.GaussianBlur(data, (k, k), sigma, dst=data)
                class_probs = impl.contiguous(data.T)

            if soften > 1:
                class_probs = impl.softmax(class_probs, axis=0)

        dims = tuple(class_idx.shape)

        kw_heat = {
            'class_idx': class_idx,
            'classes': classes,
            'img_dims': img_dims,
            'tf_data_to_img': tf_data_to_img,
            'datakeys': ['kpts_ignore', 'class_idx'],
        }

        if 'class_probs' not in exclude:
            kw_heat['class_probs'] = class_probs

        if 'diameter' not in exclude:
            if 'size' in fcn_target:
                kw_heat['diameter'] = fcn_target['size'][[1, 0]]

        if 'offset' not in exclude:
            if 'dxdy' in fcn_target:
                kw_heat['offset'] = fcn_target['dxdy'][[1, 0]]

        if 'keypoints' not in exclude:
            if 'kpts' in fcn_target:
                kp_classes = self.meta['kp_classes']
                K = len(kp_classes)
                # TODO: add noise or do some bluring?
                kw_heat['keypoints'] = impl.view(fcn_target['kpts'], (2, K,) + dims)[[1, 0]]
                kw_heat['kpts_ignore'] = fcn_target['kpts_ignore']

        self = kwimage.Heatmap(**kw_heat)
        # print('self.data: ' + ub.urepr(ub.map_vals(lambda x: x.shape, self.data), nl=1))
        return self


class Detections(ub.NiceRepr, _DetAlgoMixin, _DetDrawMixin):
    """
    Container for holding and manipulating multiple detections.

    Attributes:
        data (Dict): dictionary containing corresponding lists. The length of
            each list is the number of detections. This contains the bounding
            boxes, confidence scores, and class indices. Details of the most
            common keys and types are as follows:

            boxes (kwimage.Boxes[ArrayLike]): multiple bounding boxes
            scores (ArrayLike): associated scores
            class_idxs (ArrayLike): associated class indices
            segmentations (ArrayLike): segmentations masks for each box,
            members can be :class:`Mask` or :class:`MultiPolygon`.
            keypoints (ArrayLike): keypoints for each box. Members should
            be :class:`Points`.

            Additional custom keys may be specified as long as (a) the values
            are array-like and the first axis corresponds to the standard data
            values and (b) are custom keys are listed in the `datakeys` kwargs
            when constructing the Detections.

        meta (Dict):
            This contains contextual information about the detections.  This
            includes the class names, which can be indexed into via the class
            indexes.

    Example:
        >>> import kwimage
        >>> dets = kwimage.Detections(
        >>>     # there are expected keys that do not need registration
        >>>     boxes=kwimage.Boxes.random(3),
        >>>     class_idxs=[0, 1, 1],
        >>>     classes=['a', 'b'],
        >>>     # custom data attrs must align with boxes
        >>>     myattr1=np.random.rand(3),
        >>>     myattr2=np.random.rand(3, 2, 8),
        >>>     # there are no restrictions on metadata
        >>>     mymeta='a custom metadata string',
        >>>     # Note that any key not in kwimage.Detections.__datakeys__ or
        >>>     # kwimage.Detections.__metakeys__ must be registered at the
        >>>     # time of construction.
        >>>     datakeys=['myattr1', 'myattr2'],
        >>>     metakeys=['mymeta'],
        >>>     checks=True,
        >>> )
        >>> print('dets = {}'.format(dets))
        dets = <Detections(3)>
    """
    # __slots__ = ('data', 'meta',)

    # Valid keys for the data dictionary
    # NOTE: I'm not sure its productive to restrict to a set of specified
    # properties. It might be better to allow detections to have arbitrary data
    # properties like: velocity, as long as they are array-like. However, I'm
    # not sure how to best structure the code to allow this so it is both clear
    # and efficient. Currently I've allowed the user to specify custom datakeys
    # and metakeys as kwargs, but that design might change.
    __datakeys__ = ['boxes', 'scores', 'class_idxs', 'probs', 'weights',
                    'keypoints', 'segmentations']

    # Valid keys for the meta dictionary
    __metakeys__ = ['classes']

    def __init__(self, data=None, meta=None, datakeys=None, metakeys=None,
                 checks=True, **kwargs):
        """
        Construct a Detections object by either explicitly specifying the
        internal data and meta dictionary structures or by passing expected
        attribute names as kwargs.

        Args:
            data (Dict[str, ArrayLike]): explicitly specify the data dictionary
            meta (Dict[str, object]): explicitly specify the meta dictionary
            datakeys (List[str]): a list of custom attributes that should be
               considered as data (i.e. must be an array aligned with boxes).
            metakeys (List[str]): a list of custom attributes that should be
               considered as metadata (i.e. can be arbitrary).
            checks (bool): if True and arguments are passed by
                kwargs, then check / ensure that all types are compatible.
                Defaults to True.
            **kwargs:
                specify any key for the data or meta dictionaries.

        Note:
            Custom data and metadata can be specified as long as you pass the
            names of these keys in the `datakeys` and/or `metakeys` kwargs.

            In the case where you specify a custom attribute as a list, it will
            "currently" (we may change this behavior in the future) be coerced
            into a numpy or torch array. If you want to store a generic Python
            list, wrap the custom list in a ``_generic.ObjectList``.

        Example:
            >>> # Coerce to numpy
            >>> import kwimage
            >>> dets = Detections(
            >>>     boxes=kwimage.Boxes.random(3).numpy(),
            >>>     class_idxs=[0, 1, 1],
            >>>     checks=True,
            >>> )
            >>> # xdoctest: +REQUIRES(module:torch)
            >>> # Coerce to tensor
            >>> dets = Detections(
            >>>     boxes=kwimage.Boxes.random(3).tensor(),
            >>>     class_idxs=[0, 1, 1],
            >>>     checks=True,
            >>> )
            >>> # Error on incompatible types
            >>> import pytest
            >>> with pytest.raises(TypeError):
            >>>     dets = Detections(
            >>>         boxes=kwimage.Boxes.random(3).tensor(),
            >>>         scores=np.random.rand(3),
            >>>         class_idxs=[0, 1, 1],
            >>>         checks=True,
            >>>     )

        Example:
            >>> self = Detections.random(10)
            >>> other = Detections(self)
            >>> assert other.data == self.data
            >>> assert other.data is self.data, 'try not to copy unless necessary'

        """
        torch = sys.modules.get('torch', None)
        # Standardize input format
        if kwargs:
            if data or meta:
                raise ValueError('Cannot specify kwargs AND data/meta dicts')
            _datakeys = self.__datakeys__
            _metakeys = self.__metakeys__
            # Allow the user to specify custom data and meta keys
            if datakeys is not None:
                _datakeys = _datakeys + list(datakeys)
            if metakeys is not None:
                _metakeys = _metakeys + list(metakeys)
            # Perform input checks whenever kwargs is given
            data = {key: kwargs.pop(key) for key in _datakeys if key in kwargs}
            meta = {key: kwargs.pop(key) for key in _metakeys if key in kwargs}
            if kwargs:
                raise ValueError(
                    'Unknown kwargs: {}'.format(sorted(kwargs.keys())))

            if checks:
                import kwarray
                # Check to make sure all types in `data` are compatible
                ndarrays = []
                tensors = []
                other = []
                objlist = []

                ### Make it easier to specify keypoints and segmentations
                if 'segmentations' in data:
                    import kwimage
                    segmentations = data.get('segmentations', None)
                    if segmentations is not None:
                        segmentations = kwimage.SegmentationList.coerce(segmentations)
                        data['segmentations'] = segmentations

                for k, v in data.items():
                    if v is None:
                        objlist.append(v)
                    elif _generic._isinstance2(v, _generic.ObjectList):
                        objlist.append(v)
                    elif _generic._isinstance2(v, _boxes.Boxes):
                        if v.is_numpy():
                            ndarrays.append(k)
                        else:
                            tensors.append(k)
                    elif isinstance(v, np.ndarray):
                        ndarrays.append(k)
                    elif torch is not None and isinstance(v, torch.Tensor):
                        tensors.append(k)
                    else:
                        other.append(k)

                if bool(ndarrays) and bool(tensors):
                    raise TypeError(
                        'Detections can hold numpy.ndarrays or torch.Tensors, '
                        'but not both')
                if tensors:
                    impl = kwarray.ArrayAPI.coerce('tensor')
                else:
                    impl = kwarray.ArrayAPI.coerce('numpy')
                for k in other:
                    data[k] = impl.asarray(data[k])

        elif isinstance(data, self.__class__):
            # Avoid runtime checks and assume the user is doing the right thing
            # if data and meta are explicitly specified
            meta = data.meta
            data = data.data
        if meta is None:
            meta = {}

        self.data = data
        self.meta = meta

    def __nice__(self):
        return self.num_boxes()

    def __len__(self):
        return self.num_boxes()

    def copy(self):
        """
        Returns a deep copy of this Detections object
        """
        import copy
        return copy.deepcopy(self)

    @classmethod
    def coerce(cls, data=None, **kwargs):
        """
        The "try-anything to get what I want" constructor

        Args:
            data:
            **kwargs: currently boxes and cnames

        Example:
            >>> from kwimage.structs.detections import *  # NOQA
            >>> import kwimage
            >>> kwargs = dict(
            >>>     boxes=kwimage.Boxes.random(4),
            >>>     cnames=['a', 'b', 'c', 'c'],
            >>> )
            >>> data = {}
            >>> self = kwimage.Detections.coerce(data, **kwargs)
        """
        if data is None:
            data = {}
        if 'boxes' in kwargs:
            data['boxes'] = kwargs['boxes']

        cnames = kwargs.get('cnames', kwargs.get('class_names', kwargs.get('catnames', None)))
        if cnames is not None:
            if len(cnames) and isinstance(ub.peek(cnames), str):
                if 'classes' not in data:
                    data['classes'] = sorted(set(cnames))
                if 'class_idxs' not in data:
                    classes = data['classes']
                    data['class_idxs'] = list(map(classes.index, cnames))

        self = cls(**data)
        return self

    @classmethod
    def from_coco_annots(cls, anns, cats=None, classes=None, kp_classes=None,
                         shape=None, dset=None):
        """
        Create a Detections object from a list of coco-like annotations.

        Args:
            anns (List[Dict]): list of coco-like annotation objects

            dset (kwcoco.CocoDataset): if specified, cats, classes, and kp_classes
                can are ignored.

            cats (List[Dict]): coco-format category information.
                Used only if `dset` is not specified.

            classes (kwcoco.CategoryTree): category tree with coco class
                info. Used only if `dset` is not specified.

            kp_classes (kwcoco.CategoryTree): keypoint category tree with
                coco keypoint class info. Used only if `dset` is not specified.

            shape (tuple): shape of parent image

        Returns:
            Detections: a detections object

        Example:
            >>> from kwimage.structs.detections import *  # NOQA
            >>> # xdoctest: +REQUIRES(--module:ndsampler)
            >>> anns = [{
            >>>     'id': 0,
            >>>     'image_id': 1,
            >>>     'category_id': 2,
            >>>     'bbox': [2, 3, 10, 10],
            >>>     'keypoints': [4.5, 4.5, 2],
            >>>     'segmentation': {
            >>>         'counts': '_11a04M2O0O20N101N3L_5',
            >>>         'size': [20, 20],
            >>>     },
            >>> }]
            >>> dataset = {
            >>>     'images': [],
            >>>     'annotations': [],
            >>>     'categories': [
            >>>         {'id': 0, 'name': 'background'},
            >>>         {'id': 2, 'name': 'class1', 'keypoints': ['spot']}
            >>>     ]
            >>> }
            >>> #import ndsampler
            >>> #dset = ndsampler.CocoDataset(dataset)
            >>> cats = dataset['categories']
            >>> dets = Detections.from_coco_annots(anns, cats)

        Example:
            >>> # xdoctest: +REQUIRES(--module:ndsampler)
            >>> # Test case with no category information
            >>> from kwimage.structs.detections import *  # NOQA
            >>> anns = [{
            >>>     'id': 0,
            >>>     'image_id': 1,
            >>>     'category_id': None,
            >>>     'bbox': [2, 3, 10, 10],
            >>>     'prob': [.1, .9],
            >>> }]
            >>> cats = [
            >>>     {'id': 0, 'name': 'background'},
            >>>     {'id': 2, 'name': 'class1'}
            >>> ]
            >>> dets = Detections.from_coco_annots(anns, cats)

        Example:
            >>> import kwimage
            >>> # xdoctest: +REQUIRES(--module:ndsampler)
            >>> import ndsampler
            >>> sampler = ndsampler.CocoSampler.demo('photos')
            >>> iminfo, anns = sampler.load_image_with_annots(1)
            >>> shape = iminfo['imdata'].shape[0:2]
            >>> kp_classes = sampler.dset.keypoint_categories()
            >>> dets = kwimage.Detections.from_coco_annots(
            >>>     anns, sampler.dset.dataset['categories'], sampler.catgraph,
            >>>     kp_classes, shape=shape)

        Example:
            >>> from kwimage.structs.detections import *  # NOQA
            >>> # xdoctest: +REQUIRES(--module:ndsampler)
            >>> kpclasses = kwcoco.CategoryTree.from_coco([
            >>>     {'name': 'head', 'id': 1},
            >>>     {'name': 'backbone', 'id': 2},
            >>>     {'name': 'stem', 'id': 3},
            >>> ])
            >>> points = kwimage.Points.random(3, classes=kpclasses)
            >>> anns = [{
            >>>     'id': 0,
            >>>     'image_id': 1,
            >>>     'category_id': 2,
            >>>     'bbox': [2, 3, 10, 10],
            >>>     'keypoints': points.to_coco('new-v2'),
            >>> }]
            >>> dataset = {
            >>>     'images': [],
            >>>     'annotations': anns,
            >>>     'categories': [
            >>>         {'id': 0, 'name': 'background'},
            >>>         {'id': 2, 'name': 'class1', 'keypoint_skeleton': [('head', 'backbone'), ('backbone', 'stem')]}
            >>>     ],
            >>>     'keypoint_categories': list(kpclasses.to_coco()),
            >>> }
            >>> #import ndsampler
            >>> #dset = ndsampler.CocoDataset(dataset)
            >>> cats = dataset['categories']
            >>> dets = Detections.from_coco_annots(anns, cats, kp_classes=kpclasses)
            >>> # TODO: draw a skeleton
            >>> dets.data['keypoints'].data[0].meta
        """
        import kwimage
        cnames = None
        if dset is not None:
            try:
                classes = dset.object_categories()
            except Exception:
                pass
            cats = dset.dataset['categories']  # redundant with classes, can simplify
            try:
                kp_classes = dset.keypoint_categories()
            except Exception:
                pass
            if len(kp_classes) == 0:
                kp_classes = None
        else:
            if cats is None:
                cnames = []
                for ann in anns:
                    if 'category_name' in ann:
                        cnames.append(ann['category_name'])
                    else:
                        raise Exception('Specify dset or cats or category_name in each annotation')
                if classes is None:
                    classes = sorted(set(cnames))
                assert set(cnames).issubset(set(classes))

                # make dummy cats
                cats = [{'name': name, 'id': cid}
                        for cid, name in enumerate(classes, start=1) ]

        if classes is None:
            classes = list(ub.oset([cat['name'] for cat in cats]))

        if cnames is None:
            cids = [ann['category_id'] for ann in anns]
            cid_to_cat = {c['id']: c for c in cats}  # Hack
            cnames = [None if cid is None else cid_to_cat[cid]['name']
                      for cid in cids]

        xywh = np.array([ann['bbox'] for ann in anns], dtype=np.float32)
        boxes = kwimage.Boxes(xywh, 'xywh')
        try:
            class_idxs = [classes.index(cname) for cname in cnames]
        except (KeyError, ValueError):
            class_idxs = [None if cname is None else classes.index(cname)
                          for cname in cnames]

        dets = Detections(
            boxes=boxes,
            class_idxs=np.array(class_idxs),
            classes=classes,
        )

        if len(anns):
            if 'score' in anns[0]:
                dets.data['scores'] = np.array([ann.get('score', np.nan) for ann in anns])

            if 'prob' in anns[0]:
                dets.data['probs'] = np.array([ann.get('prob', np.nan) for ann in anns])

            if 'weight' in anns[0]:
                dets.data['weights'] = np.array([ann.get('weight', np.nan) for ann in anns])

        if True:
            ss = [ann.get('segmentation', None) for ann in anns]
            masks = [
                None if s is None else
                kwimage.MultiPolygon.coerce(s, dims=shape)
                for s in ss
            ]
            dets.data['segmentations'] = kwimage.PolygonList(masks)

        HANDLE_KEYPOINTS = 1
        if HANDLE_KEYPOINTS:
            # TODO: loop a skeleton for the Points object.

            name_to_cat = {c['name']: c for c in cats}
            def _lookup_kp_class_idxs(cid):
                kpnames = None
                while kpnames is None:
                    cat = cid_to_cat[cid]
                    parent = cat.get('supercategory', None)
                    if 'keypoints' in cat:
                        kpnames = cat['keypoints']
                    elif parent is not None:
                        cid = name_to_cat[cat['supercategory']]['id']
                    else:
                        raise KeyError(cid)
                kpcidxs = [kp_classes.index(n) for n in kpnames]
                return kpcidxs
            kpts = []
            for ann in anns:
                k = ann.get('keypoints', None)
                if k is None:
                    kpts.append(k)
                elif len(k) == 0:
                    kpcidxs = []
                    kpts.append(None)
                else:
                    kpcidxs = None
                    # TODO: correctly handle newstyle keypoints
                    if dset is not None:
                        pass
                    kpcidxs = None
                    is_oldstyle = (isinstance(k, list) and len(k) and not isinstance(ub.peek(k), dict))
                    if is_oldstyle:
                        # oldstyle
                        if kp_classes is not None:
                            # These are only needed for old-style coco
                            cid = ann['category_id']
                            try:
                                kpcidxs = _lookup_kp_class_idxs(cid)
                            except KeyError:
                                # unable to find keypoint category ids
                                ...

                    pts = kwimage.Points.from_coco(
                        k, class_idxs=kpcidxs, classes=kp_classes)
                    kpts.append(pts)

            if __debug__:
                boxes = dets.data.get('boxes', None)
                if boxes is not None:
                    assert len(kpts) == len(boxes)

            dets.data['keypoints'] = kwimage.PointsList(kpts)

            if kp_classes is not None:
                dets.data['keypoints'].meta['classes'] = kp_classes
                dets.meta['kp_classes'] = kp_classes
        return dets

    def to_coco(self, cname_to_cat=None, style='orig', image_id=None, dset=None):
        """
        Converts this set of detections into coco-like annotation dictionaries.

        Note:
            Not all aspects of the MS-COCO format can be accurately
            represented, so some liberties are taken. The MS-COCO standard
            defines that annotations should specifiy a category_id field, but
            in some cases this information is not available so we will populate
            a 'category_name' field if possible and in the worst case fall back
            to 'category_index'.

            Additionally, detections may contain additional information beyond
            the MS-COCO standard, and this information (e.g. weight, prob,
            score) is added as forign fields.

        Args:
            cname_to_cat: currently ignored.

            style (str): either 'orig' (for the original coco
                format) or 'new' for the more general kwcoco-style coco
                format. Defaults to 'orig'

            image_id (int):
                if specified, populates the image_id field of each image.

            dset (kwcoco.CocoDataset | None):
                if specified, attempts to populate the category_id field
                to be compatible with this coco dataset.

        Yields:
            dict: coco-like annotation structures

        Example:
            >>> # xdoctest: +REQUIRES(module:ndsampler)
            >>> from kwimage.structs.detections import *
            >>> self = Detections.demo()[0]
            >>> cname_to_cat = None
            >>> list(self.to_coco())
        """
        import kwarray
        to_collate = {}
        boxes = self.data.get('boxes', None)
        if boxes is not None:
            to_collate['bbox'] = list(boxes.to_coco(style=style))

        class_idxs = self.data.get('class_idxs', None)
        if class_idxs is not None:
            if 'classes' in self.meta:
                classes = self.meta['classes']
                catnames = [classes[cidx] for cidx in class_idxs]
                if cname_to_cat is not None:
                    pass
                if dset is not None:
                    cids = [dset._resolve_to_cat(c)['id'] for c in catnames]
                    to_collate['category_id'] = cids
                else:
                    to_collate['category_name'] = catnames
            else:
                if dset is not None:
                    raise NotImplementedError(
                        'Passed a dset to resolve category id, but this '
                        'detection object has no classes meta attribute')
                to_collate['category_index'] = kwarray.ArrayAPI.tolist(class_idxs)

        keypoints = self.data.get('keypoints', None)
        if keypoints is not None:
            to_collate['keypoints'] = list(keypoints.to_coco(style=style))

        segmentations = self.data.get('segmentations', None)
        if segmentations is not None:
            to_collate['segmentation'] = list(segmentations.to_coco(style=style))

        scores = self.data.get('scores', None)
        if scores is not None:
            to_collate['score'] = kwarray.ArrayAPI.tolist(scores)

        weights = self.data.get('weights', None)
        if weights is not None:
            to_collate['weight'] = kwarray.ArrayAPI.tolist(weights)

        probs = self.data.get('probs', None)
        if probs is not None:
            to_collate['prob'] = kwarray.ArrayAPI.tolist(probs)

        if image_id is not None:
            to_collate['image_id'] = [image_id] * len(self)

        if __debug__:
            # Error before we silently return bad data
            collate_lens = ub.udict(to_collate).map_values(len)
            assert ub.allsame(collate_lens.values())

        keys = list(to_collate.keys())
        for item_vals in zip(*to_collate.values()):
            ann = ub.dzip(keys, item_vals)
            yield ann

    # --- Data Properties ---

    @property
    def boxes(self):
        return self.data['boxes']

    @property
    def class_idxs(self):
        return self.data['class_idxs']

    @property
    def scores(self):
        """ typically only populated for predicted detections """
        return self.data['scores']

    @property
    def probs(self):
        """ typically only populated for predicted detections """
        return self.data['probs']

    @property
    def weights(self):
        """ typically only populated for groundtruth detections """
        return self.data['weights']

    # --- Meta Properties ---

    @property
    def classes(self):
        return self.meta.get('classes', None)

    def num_boxes(self):
        return len(self.boxes)

    # --- Modifiers ---

    def warp(self, transform, input_dims=None, output_dims=None, inplace=False):
        """
        Spatially warp the detections.

        Args:
            transform (kwimage.Affine | ndarray | Callable | Any):
                Something coercable to a transform.
                Usually a kwimage.Affine object

            input_dims (Tuple[int, int]): shape of the expected input canvas

            output_dims (Tuple[int, int]): shape of the expected output canvas

            inplace (bool): if true operate inplace

        Returns:
            Detections: the warped detections object

        Example:
            >>> import skimage
            >>> transform = skimage.transform.AffineTransform(scale=(2, 3), translation=(4, 5))
            >>> self = Detections.random(2)
            >>> new = self.warp(transform)
            >>> assert new.boxes == self.boxes.warp(transform)
            >>> assert new != self
        """
        new = self if inplace else self.__class__(self.data.copy(), self.meta)
        new.data['boxes'] = new.data['boxes'].warp(transform,
                                                   input_dims=input_dims,
                                                   inplace=inplace)
        if new.data.get('keypoints', None) is not None:
            new.data['keypoints'] = new.data['keypoints'].warp(
                transform, input_dims=input_dims, output_dims=output_dims,
                inplace=inplace)
        if new.data.get('segmentations', None) is not None:
            new.data['segmentations'] = new.data['segmentations'].warp(
                transform, input_dims=input_dims, output_dims=output_dims,
                inplace=inplace)
        return new

    def scale(self, factor, output_dims=None, inplace=False):
        """
        Spatially scale the detections.

        Example:
            >>> import skimage
            >>> transform = skimage.transform.AffineTransform(scale=(2, 3), translation=(4, 5))
            >>> self = Detections.random(2)
            >>> new = self.warp(transform)
            >>> assert new.boxes == self.boxes.warp(transform)
            >>> assert new != self
        """
        new = self if inplace else self.__class__(self.data.copy(), self.meta)
        new.data['boxes'] = new.data['boxes'].scale(factor, inplace=inplace)
        if new.data.get('keypoints', None) is not None:
            new.data['keypoints'] = new.data['keypoints'].scale(
                factor, output_dims=output_dims, inplace=inplace)
        if new.data.get('segmentations', None) is not None:
            new.data['segmentations'] = new.data['segmentations'].scale(
                factor, output_dims=output_dims, inplace=inplace)
        return new

    def translate(self, offset, output_dims=None, inplace=False):
        """
        Spatially translate the detections.

        Example:
            >>> import skimage
            >>> self = Detections.random(2)
            >>> new = self.translate(10)
        """
        new = self if inplace else self.__class__(self.data.copy(), self.meta)
        new.data['boxes'] = new.data['boxes'].translate(offset, inplace=inplace)
        if new.data.get('keypoints', None) is not None:
            new.data['keypoints'] = new.data['keypoints'].translate(
                offset, output_dims=output_dims)
        if new.data.get('segmentations', None) is not None:
            new.data['segmentations'] = new.data['segmentations'].translate(
                offset, output_dims=output_dims)
        return new

    @classmethod
    def concatenate(cls, dets):
        """
        Args:
            boxes (Sequence[Detections]): list of detections to concatenate

        Returns:
            Detections: stacked detections

        Example:
            >>> self = Detections.random(2)
            >>> other = Detections.random(3)
            >>> dets = [self, other]
            >>> new = Detections.concatenate(dets)
            >>> assert new.num_boxes() == 5

            >>> self = Detections.random(2, segmentations=True)
            >>> other = Detections.random(3, segmentations=True)
            >>> dets = [self, other]
            >>> new = Detections.concatenate(dets)
            >>> assert new.num_boxes() == 5
        """
        if len(dets) == 0:
            raise ValueError('need at least one detection to concatenate')
        newdata = {}
        first = dets[0]
        for key in first.data.keys():
            if first.data[key] is None:
                newdata[key] = None
            else:
                try:
                    tocat = [d.data[key] for d in dets]
                    try:
                        # Use class concatenate if it exists,
                        cat = tocat[0].__class__.concatenate
                    except AttributeError:
                        # otherwise use numpy/torch
                        cat = _boxes._cat
                    newdata[key] = cat(tocat, axis=0)
                except Exception:
                    msg = ('Error when trying to concat {}'.format(key))
                    print(msg)
                    raise

        newmeta = dets[0].meta
        new = cls(newdata, newmeta)
        return new

    def argsort(self, reverse=True):
        """
        Sorts detection indices by descending (or ascending) scores

        Returns:
            ndarray[Shape['*'], Integer]: sorted indices
            torch.Tensor: sorted indices if using torch backends
        """
        sortx = self.scores.argsort()
        if reverse:
            torch = sys.modules.get('torch', None)
            if torch is not None and torch.is_tensor(sortx):
                sortx = torch.flip(sortx, dims=(0,))
            else:
                sortx = sortx[::-1]
        return sortx

    def sort(self, reverse=True):
        """
        Sorts detections by descending (or ascending) scores

        Returns:
            kwimage.structs.Detections: sorted copy of self
        """
        sortx = self.argsort(reverse=reverse)
        return self.take(sortx)

    def compress(self, flags, axis=0):
        """
        Returns a subset where corresponding locations are True.

        Args:
            flags (ndarray[Any, Bool] | torch.Tensor):
                mask marking selected items

        Returns:
            kwimage.structs.Detections: subset of self

        CommandLine:
            xdoctest -m kwimage.structs.detections Detections.compress

        Example:
            >>> # xdoctest: +REQUIRES(module:torch)
            >>> import kwimage
            >>> dets = kwimage.Detections.random(keypoints='dense')
            >>> flags = np.random.rand(len(dets)) > 0.5
            >>> subset = dets.compress(flags)
            >>> assert len(subset) == flags.sum()
            >>> subset = dets.tensor().compress(flags)
            >>> assert len(subset) == flags.sum()
        """
        if flags is Ellipsis:
            return self

        if len(flags) != len(self):
            raise IndexError('compress must get a flag for every item')

        if self.is_tensor():
            torch = sys.modules.get('torch', None)
            if isinstance(flags, np.ndarray):
                if flags.dtype.kind == 'b':
                    flags = flags.astype(np.uint8)
            if isinstance(flags, torch.Tensor):
                if flags.dtype != torch.bool:
                    flags = flags.bool()
                if flags.device != flags.device:
                    flags = flags.to(self.device)
            else:
                flags = torch.BoolTensor(flags).to(self.device)
        newdata = {k: _generic._safe_compress(v, flags, axis)
                   for k, v in self.data.items()}
        return self.__class__(newdata, self.meta)

    def take(self, indices, axis=0):
        """
        Returns a subset specified by indices

        Args:
            indices (ndarray[Any, Integer]): indices to select

        Returns:
            kwimage.structs.Detections: subset of self

        Example:
            >>> import kwimage
            >>> dets = kwimage.Detections(boxes=kwimage.Boxes.random(10))
            >>> subset = dets.take([2, 3, 5, 7])
            >>> assert len(subset) == 4
            >>> # xdoctest: +REQUIRES(module:torch)
            >>> subset = dets.tensor().take([2, 3, 5, 7])
            >>> assert len(subset) == 4
        """
        if self.is_tensor():
            torch = sys.modules.get('torch', None)
            indices = torch.LongTensor(indices).to(self.device)
        newdata = {k: _generic._safe_take(v, indices, axis)
                   for k, v in self.data.items()}
        return self.__class__(newdata, self.meta)

    def __getitem__(self, index):
        """
        Fancy slicing / subset / indexing.

        Note: scalar indices are always coerced into index lists of length 1.

        Example:
            >>> import kwimage
            >>> import kwarray
            >>> dets = kwimage.Detections(boxes=kwimage.Boxes.random(10))
            >>> indices = [2, 3, 5, 7]
            >>> flags = kwarray.boolmask(indices, len(dets))
            >>> assert dets[flags].data == dets[indices].data
        """
        if isinstance(index, slice):
            index = list(range(*index.indices(len(self))))
        if ub.iterable(index):
            import kwarray
            impl = kwarray.ArrayAPI.coerce('numpy')
            indices = impl.asarray(index)
        else:
            indices = np.array([index])
        if indices.dtype.kind == 'b':
            return self.compress(indices)
        else:
            return self.take(indices)

    @property
    def device(self):
        """ If the backend is torch returns the data device, otherwise None """
        return self.boxes.device

    def is_tensor(self):
        """ is the backend fueled by torch? """
        return self.boxes.is_tensor()

    def is_numpy(self):
        """ is the backend fueled by numpy? """
        return self.boxes.is_numpy()

    def numpy(self):
        """
        Converts tensors to numpy. Does not change memory if possible.

        Example:
            >>> # xdoctest: +REQUIRES(module:torch)
            >>> self = Detections.random(3).tensor()
            >>> newself = self.numpy()
            >>> self.scores[0] = 0
            >>> assert newself.scores[0] == 0
            >>> self.scores[0] = 1
            >>> assert self.scores[0] == 1
            >>> self.numpy().numpy()
        """
        newdata = {}
        for key, val in self.data.items():
            if val is None:
                newval = val
            else:
                torch = sys.modules.get('torch', None)
                if torch is not None and torch.is_tensor(val):
                    newval = val.data.cpu().numpy()
                elif hasattr(val, 'numpy'):
                    newval = val.numpy()
                else:
                    newval = val
            newdata[key] = newval
        newself = self.__class__(newdata, self.meta)
        return newself

    @property
    def dtype(self):
        dtypes = set()
        for key, val in self.data.items():
            if val is not None:
                try:
                    child_dtype = val.dtype
                    if isinstance(child_dtype, set):
                        dtypes.update(child_dtype)
                    else:
                        dtypes.add(child_dtype)
                except AttributeError:
                    dtypes.add('unknown-for-{}'.format(type(val)))
        if len(dtypes) == 1:
            return ub.peek(dtypes)
        else:
            return dtypes

    def tensor(self, device=ub.NoParam):
        """
        Converts numpy to tensors. Does not change memory if possible.

        Example:
            >>> # xdoctest: +REQUIRES(module:torch)
            >>> from kwimage.structs.detections import *
            >>> self = Detections.random(3)
            >>> newself = self.tensor()
            >>> self.scores[0] = 0
            >>> assert newself.scores[0] == 0
            >>> self.scores[0] = 1
            >>> assert self.scores[0] == 1
            >>> self.tensor().tensor()
        """
        newdata = {}
        for key, val in self.data.items():
            if val is None:
                newval = val
            elif hasattr(val, 'tensor'):
                newval = val.tensor(device)
            else:
                torch = sys.modules.get('torch', None)
                if torch is not None and torch.is_tensor(val):
                    newval = val
                else:
                    newval = torch.from_numpy(val)
                if device is not ub.NoParam:
                    newval = newval.to(device)
            newdata[key] = newval
        newself = self.__class__(newdata, self.meta)
        return newself

    # --- Non-core methods ----

    @classmethod
    def demo(Detections):
        import ndsampler
        sampler = ndsampler.CocoSampler.demo('photos')
        iminfo, anns = sampler.load_image_with_annots(1)
        input_dims = iminfo['imdata'].shape[0:2]
        kp_classes = sampler.dset.keypoint_categories()
        self = Detections.from_coco_annots(
            anns, sampler.dset.dataset['categories'],
            sampler.catgraph, kp_classes, shape=input_dims)

        # TODO: should this extra info belong in the metadata field?
        return self, iminfo, sampler

    @classmethod
    def random(cls, num=10, scale=1.0, classes=3, keypoints=False,
               segmentations=False, tensor=False, rng=None):
        """
        Creates dummy data, suitable for use in tests and benchmarks

        Args:
            num (int): number of boxes
            scale (float | tuple): bounding image size. Defaults to 1.0
            classes (int | Sequence): list of class labels or number of classes
            keypoints (bool):
                if True include random keypoints for each box.
                Defaults to False.
            segmentations (bool):
                if True include random segmentations for each box.
                Defaults to False.
            tensor (bool): determines backend.
                DEPRECATED.  Call ``.tensor()`` on resulting object instead.
            rng (int | RandomState | None):
                random state or seed

        Returns:
            Detections: random detections

        Example:
            >>> import kwimage
            >>> dets = kwimage.Detections.random(keypoints='jagged')
            >>> dets.data['keypoints'].data[0].data
            >>> dets.data['keypoints'].meta
            >>> dets = kwimage.Detections.random(keypoints='dense')
            >>> dets = kwimage.Detections.random(keypoints='dense', segmentations=True).scale(1000)
            >>> # xdoctest:+REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.autompl()
            >>> dets.draw(setlim=True)

        Example:
            >>> import kwimage
            >>> dets = kwimage.Detections.random(
            >>>     keypoints='jagged', segmentations=True, rng=0).scale(1000)
            >>> print('dets = {}'.format(dets))
            dets = <Detections(10)>
            >>> dets.data['boxes'].quantize(inplace=True)
            >>> print('dets.data = {}'.format(ub.urepr(
            >>>     dets.data, nl=1, with_dtype=False, strvals=True, sort=1)))
            dets.data = {
                'boxes': <Boxes(xywh,
                             array([[548, 544,  55, 172],
                                    [423, 645,  15, 247],
                                    [791, 383, 173, 146],
                                    [ 71,  87, 498, 839],
                                    [ 20, 832, 759,  39],
                                    [461, 780, 518,  20],
                                    [118, 639,  26, 306],
                                    [264, 414, 258, 361],
                                    [ 18, 568, 439,  50],
                                    [612, 616, 332,  66]]...))>,
                'class_idxs': [1, 2, 0, 0, 2, 0, 0, 0, 0, 0],
                'keypoints': <PointsList(n=10)>,
                'scores': [0.3595079 , 0.43703195, 0.6976312 , 0.06022547, 0.66676672, 0.67063787,0.21038256, 0.1289263 , 0.31542835, 0.36371077],
                'segmentations': <SegmentationList(n=10)>,
            }
            >>> # xdoctest:+REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.autompl()
            >>> dets.draw(setlim=True)

        Example:
            >>> # Boxes position/shape within 0-1 space should be uniform.
            >>> # xdoctest: +REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.autompl()
            >>> fig = kwplot.figure(fnum=1, doclf=True)
            >>> fig.gca().set_xlim(0, 128)
            >>> fig.gca().set_ylim(0, 128)
            >>> import kwimage
            >>> kwimage.Detections.random(num=10, segmentations=True).scale(128).draw()
        """
        import kwimage
        import kwarray
        rng = kwarray.ensure_rng(rng)
        boxes = kwimage.Boxes.random(num=num, rng=rng)
        if isinstance(classes, int):
            num_classes = classes
            classes = ['class_{}'.format(c) for c in range(classes)]
            # hack: ensure that we have a background class
            classes.append('background')
        else:
            num_classes = len(classes)
        scores = rng.rand(len(boxes))
        class_idxs = rng.randint(0, num_classes, size=len(boxes))
        self = cls(boxes=boxes, scores=scores, class_idxs=class_idxs,
                   classes=classes)
        self.meta['classes'] = classes

        if keypoints is True:
            keypoints = 'jagged'

        if segmentations:
            sseg_list = []
            for xywh in self.boxes.to_xywh().data:
                box_scale = xywh[2:]
                box_offset = xywh[0:2]
                sseg = kwimage.MultiPolygon.random(n=1, tight=True, rng=rng)
                sseg = sseg.scale(box_scale).translate(box_offset)
                sseg_list.append(sseg)
            self.data['segmentations'] = kwimage.SegmentationList.coerce(sseg_list)

        if isinstance(keypoints, str):
            kp_classes = [1, 2, 3, 4]
            self.meta['kp_classes'] = kp_classes
            if keypoints == 'jagged':
                kpts_list = kwimage.PointsList([
                    kwimage.Points.random(
                        num=rng.randint(len(kp_classes)),
                        classes=kp_classes, rng=rng,
                    )
                    for _ in range(len(boxes))
                ])
                kpts_list.meta['classes'] = kp_classes
                self.data['keypoints'] = kpts_list
            elif keypoints == 'dense':
                keypoints = kwimage.Points.random(
                    num=(len(boxes), len(kp_classes)), rng=rng,
                    classes=kp_classes,)
                self.data['keypoints'] = keypoints

        self = self.scale(scale)

        if tensor:
            ub.schedule_deprecation(
                'kwimage', 'tensor', 'argument to Detections.random',
                migration='use .tensor() instead',
                deprecate='0.8.0', error='1.0.0', remove='1.1.0'
            )
            self = self.tensor()

        return self


def _dets_to_fcmaps(dets, bg_size, input_dims, bg_idx=0, pmin=0.6, pmax=1.0,
                    soft=True, exclude=[]):
    """
    Construct semantic segmentation detection targets from annotations in
    dictionary format.

    Rasterize detections.

    Args:
        dets (kwimage.Detections):
        bg_size (tuple): size (W, H) to predict for backgrounds
        input_dims (tuple): window H, W

    Returns:
        dict: with keys
            size : 2D ndarray containing the W,H of the object
            dxdy : 2D ndarray containing the x,y offset of the object
            cidx : 2D ndarray containing the class index of the object

    Ignore:
        import xdev
        globals().update(xdev.get_func_kwargs(_dets_to_fcmaps))

    Example:
        >>> # xdoctest: +REQUIRES(module:ndsampler)
        >>> from kwimage.structs.detections import *  # NOQA
        >>> from kwimage.structs.detections import _dets_to_fcmaps
        >>> import kwimage
        >>> import ndsampler
        >>> sampler = ndsampler.CocoSampler.demo('photos')
        >>> iminfo, anns = sampler.load_image_with_annots(1)
        >>> image = iminfo['imdata']
        >>> input_dims = image.shape[0:2]
        >>> kp_classes = sampler.dset.keypoint_categories()
        >>> dets = kwimage.Detections.from_coco_annots(
        >>>     anns, sampler.dset.dataset['categories'],
        >>>     sampler.catgraph, kp_classes, shape=input_dims)
        >>> bg_size = [100, 100]
        >>> bg_idxs = sampler.catgraph.index('background')
        >>> fcn_target = _dets_to_fcmaps(dets, bg_size, input_dims, bg_idxs)
        >>> fcn_target.keys()
        >>> print('fcn_target: ' + ub.urepr(ub.map_vals(lambda x: x.shape, fcn_target), nl=1, sort=1))
        fcn_target: {
            'cidx': (512, 512),
            'class_probs': (10, 512, 512),
            'dxdy': (2, 512, 512),
            'kpts': (2, 7, 512, 512),
            'kpts_ignore': (7, 512, 512),
            'size': (2, 512, 512),
        }
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> size_mask = fcn_target['size']
        >>> dxdy_mask = fcn_target['dxdy']
        >>> cidx_mask = fcn_target['cidx']
        >>> kpts_mask = fcn_target['kpts']
        >>> def _vizmask(dxdy_mask):
        >>>     dx, dy = dxdy_mask
        >>>     mag = np.sqrt(dx ** 2 + dy ** 2)
        >>>     mag /= (mag.max() + 1e-9)
        >>>     mask = (cidx_mask != 0).astype(np.float32)
        >>>     angle = np.arctan2(dy, dx)
        >>>     orimask = kwplot.make_orimask(angle, mask, alpha=mag)
        >>>     vecmask = kwplot.make_vector_field(
        >>>         dx, dy, stride=4, scale=0.1, thickness=1, tipLength=.2,
        >>>         line_type=16)
        >>>     return [vecmask, orimask]
        >>> vecmask, orimask = _vizmask(dxdy_mask)
        >>> raster = kwimage.overlay_alpha_layers(
        >>>     [vecmask, orimask, image], keepalpha=False)
        >>> raster = dets.draw_on((raster * 255).astype(np.uint8),
        >>>                       labels=True, alpha=None)
        >>> kwplot.imshow(raster)
        >>> kwplot.show_if_requested()

        raster = (kwimage.overlay_alpha_layers(_vizmask(kpts_mask[:, 5]) + [image], keepalpha=False) * 255).astype(np.uint8)
        kwplot.imshow(raster, pnum=(1, 3, 2), fnum=1)
        raster = (kwimage.overlay_alpha_layers(_vizmask(kpts_mask[:, 6]) + [image], keepalpha=False) * 255).astype(np.uint8)
        kwplot.imshow(raster, pnum=(1, 3, 3), fnum=1)
        raster = (kwimage.overlay_alpha_layers(_vizmask(dxdy_mask) + [image], keepalpha=False) * 255).astype(np.uint8)
        raster = dets.draw_on(raster, labels=True, alpha=None)
        kwplot.imshow(raster, pnum=(1, 3, 1), fnum=1)
        raster = kwimage.overlay_alpha_layers(
            [vecmask, orimask, image], keepalpha=False)
        raster = dets.draw_on((raster * 255).astype(np.uint8),
                              labels=True, alpha=None)
        kwplot.imshow(raster)
        kwplot.show_if_requested()
    """
    import cv2
    import kwimage
    # In soft mode we made a one-channel segmentation target mask
    cidx_mask = np.full(input_dims, dtype=np.int32, fill_value=bg_idx)

    if 'class_probs' not in exclude:
        if soft:
            # In soft mode we add per-class channel probability blips
            num_obj_classes = len(dets.classes)
            cidx_probs = np.full((num_obj_classes,) + tuple(input_dims),
                                 dtype=np.float32, fill_value=0)

    if 'diameter' not in exclude:
        size_mask = np.empty((2,) + tuple(input_dims), dtype=np.float32)
        size_mask[:] = np.array(bg_size)[:, None, None]

    if 'offset' not in exclude:
        dxdy_mask = np.zeros((2,) + tuple(input_dims), dtype=np.float32)

    dets = dets.numpy()

    cxywh = dets.boxes.to_cxywh().data
    class_idxs = dets.class_idxs

    if 'segmentations' in dets.data:
        sseg_list = [None if p is None else p.to_mask(input_dims)
                     for p in dets.data['segmentations']]
    else:
        sseg_list = [None] * len(dets)

    kpts_mask = None
    if 'keypoints' in dets.data and 'keypoints' not in exclude:
        kp_classes = None
        if 'classes' in dets.data['keypoints'].meta:
            kp_classes = dets.data['keypoints'].meta['classes']
        else:
            for kp in dets.data['keypoints']:
                if kp is not None and 'classes' in kp.meta:
                    kp_classes = kp.meta['classes']
                    break

        if kp_classes is not None:
            num_kp_classes = len(kp_classes)
            kpts_mask = np.zeros((2, num_kp_classes) + tuple(input_dims),
                                 dtype=np.float32)

        pts_list = dets.data['keypoints'].data
        for pts in pts_list:
            if pts is not None:
                pass

        kpts_ignore_mask = np.ones((num_kp_classes,) + tuple(input_dims),
                                   dtype=np.float32)
    else:
        pts_list = [None] * len(dets)

    # Overlay smaller classes on top of larger ones
    if len(cxywh):
        area = cxywh[..., 2] * cxywh[..., 2]
    else:
        area = []

    # TODO: maybe take scores into account?
    sortx = np.argsort(area)[::-1]
    cxywh = cxywh[sortx]
    class_idxs = class_idxs[sortx]
    pts_list = list(ub.take(pts_list, sortx))
    sseg_list = list(ub.take(sseg_list, sortx))

    def iround(x):
        return int(round(x))

    H, W = input_dims
    xcoord, ycoord = np.meshgrid(np.arange(W), np.arange(H))

    for box, cidx, sseg_mask, pts in zip(cxywh, class_idxs, sseg_list, pts_list):
        (cx, cy, w, h) = box
        center = (iround(cx), iround(cy))
        # Adjust so smaller objects get more pixels
        wf = min(1, (w / 64))
        hf = min(1, (h / 64))
        # wf = min(1, (w / W))
        # hf = min(1, (h / H))
        wf = (1 - wf) * pmax + wf * pmin
        hf = (1 - hf) * pmax + hf * pmin
        half_w = iround(wf * w / 2 + 1)
        half_h = iround(hf * h / 2 + 1)
        axes = (half_w, half_h)

        if sseg_mask is None:
            mask = np.zeros_like(cidx_mask, dtype=np.uint8)
            if mask.size > 0:
                mask = cv2.ellipse(mask, center, axes, angle=0.0,
                                   startAngle=0.0, endAngle=360.0, color=1,
                                   thickness=-1)
            mask = mask.astype(bool)
        else:
            mask = sseg_mask.to_c_mask().data.astype(bool)
        # class index
        cidx_mask[mask] = int(cidx)
        if 'class_probs' not in exclude:
            if soft:
                blip = kwimage.gaussian_patch((half_h * 2, half_w * 2))
                blip = blip / blip.max()
                subindex = (slice(cy - half_h, cy + half_h),
                            slice(cx - half_w, cx + half_w))
                kwimage.subpixel_maximum(cidx_probs[cidx], blip, subindex)

        # object size
        if 'diameter' not in exclude:
            size_mask[0][mask] = float(w)
            size_mask[1][mask] = float(h)

            assert np.all(size_mask[0][mask] == float(w))

        # object offset
        if 'offset' not in exclude:
            dx = cx - xcoord[mask]
            dy = cy - ycoord[mask]
            dxdy_mask[0][mask] = dx
            dxdy_mask[1][mask] = dy

        if kpts_mask is not None:
            if 'keypoints' not in exclude:
                if pts is not None:
                    # Keypoint offsets
                    _xys = pts.data['xy'].data
                    if len(_xys) > 0:
                        _cidxs = pts.data['class_idxs']
                        if _cidxs is None:
                            # import warnings
                            # warnings.warn('Cannot rasterize keypoints with undefined categories')
                            raise ValueError(
                                'cannot rasterize keypoints with undefined categories')
                        else:
                            for xy, kp_cidx in zip(_xys, _cidxs):
                                if kp_cidx < 0:
                                    import warnings
                                    warnings.warn('Cannot rasterize keypoints with unknown classes')
                                else:
                                    kp_x, kp_y = xy
                                    kp_dx = kp_x - xcoord[mask]
                                    kp_dy = kp_y - ycoord[mask]
                                    kpts_mask[0, kp_cidx][mask] = kp_dx
                                    kpts_mask[1, kp_cidx][mask] = kp_dy
                                    kpts_ignore_mask[kp_cidx][mask] = 0

    fcn_target = {
        'cidx': cidx_mask,
    }
    if 'diameter' not in exclude:
        fcn_target['size'] = size_mask

    if 'offset' not in exclude:
        fcn_target['dxdy'] = dxdy_mask

    if 'class_probs' not in exclude:
        if soft:
            nonbg_idxs = sorted(set(range(num_obj_classes)) - {bg_idx})
            cidx_probs[bg_idx] = 1 - cidx_probs[nonbg_idxs].sum(axis=0)
            fcn_target['class_probs'] = cidx_probs

    if kpts_mask is not None:
        if 'keypoints' not in exclude:
            fcn_target['kpts'] = kpts_mask
            fcn_target['kpts_ignore'] = kpts_ignore_mask
    else:
        if 'keypoints' in dets.data:
            if any(kp is not None for kp in dets.data['keypoints']):
                raise AssertionError(
                    'dets had keypoints, but we didnt encode them, were the kp classes missing?')

    return fcn_target


class _UnitDoctTests:
    """
    Hacking in unit tests as doctests the file itself so it is easy to move to
    kwannot when I finally get around to that.
    """

    def _test_foreign_keys_compress():
        """
        A detections object should be able to maintain foreign keys through
        compress operations.

        Example:
            >>> from kwimage.structs.detections import _UnitDoctTests
            >>> from kwimage.structs.detections import _generic
            >>> _UnitDoctTests._test_foreign_keys_compress()
        """
        import kwimage
        n = 5
        dets = kwimage.Detections.random(num=n)
        flags = dets.scores > np.median(dets.scores)

        # Test normal compress
        reduced = dets.compress(flags)
        m = len(reduced)

        # Test case with None attribute
        dets2 = kwimage.Detections(**{
            'boxes': dets.data['boxes'],
            'custom': None,
            'datakeys': ['custom'],
        })
        reduced2 = dets2.compress(flags)
        assert dets2.data['custom'] is None, 'should be able to specify None value'
        assert reduced2.data['custom'] is None, 'should be able to specify None value'

        # Test case with _generic.ObjectList[None] attribute
        dets3 = kwimage.Detections(**{
            'boxes': dets.data['boxes'],
            'custom': _generic.ObjectList([None] * n),
            'datakeys': ['custom'],
        })
        reduced3 = dets3.compress(flags)
        assert dets3.data['custom'].data == [None] * n, 'should be able to specify ObjectList[None] value'
        assert reduced3.data['custom'].data == [None] * m, 'should be able to specify ObjectList[None] value'
        assert len(reduced3.data['custom']) == m, 'compress failed'

        # NOTE: We expect Lists to always be coreced to arrays
        # Test case with List[None] attribute
        dets4 = kwimage.Detections(**{
            'boxes': dets.data['boxes'],
            'custom': [None] * n,
            'datakeys': ['custom'],
        })
        reduced4 = dets4.compress(flags)
        assert dets4.data['custom'].dtype.kind == 'O', (
            'we currently expect list to be coerced (may change in the future)')
        assert reduced4.data['custom'].dtype.kind == 'O', (
            'we currently expect list to be coerced (may change in the future)')
        assert len(reduced4.data['custom']) == m, 'compress failed'


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m kwimage.structs.detections
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



from _typeshed import Incomplete

KWIMAGE_DISABLE_WARNINGS: Incomplete
KWIMAGE_DISABLE_TRANSFORM_WARNINGS: Incomplete
KWIMAGE_DISABLE_IMPORT_WARNINGS: Incomplete
KWIMAGE_DISABLE_TORCHVISION_NMS: Incomplete
KWIMAGE_DISABLE_C_EXTENSIONS: Incomplete



import itertools as it
import numpy as np


def _draw_text_on_image_pil(img, text, org=None, fontpath=None, fontsize=32):
    r"""
    PIL backend.

    TODO:
        - [ ] Abstract `draw_text_on_image` to have an opencv and PIL backend.
              OpenCV can't do unicode, but PIL needs an external font file.
              However, PIL seems to have a much prettier result, we may
              want to use that as the default.

    References:
        https://stackoverflow.com/questions/50854235/how-to-draw-chinese-text-on-the-image-using-cv2-puttextcorrectly-pythonopen

    Ignore:
        from kwimage.im_draw import *  # NOQA
        from kwimage.im_draw import _draw_text_on_image_pil, _text_sizes, _broadcast_colors, _masked_checkerboard
        img = np.zeros((128, 128, 3), dtype=np.uint8)
        text = 'hello\n⬍\nworld'
        org = (10, 10)
        new_img = _draw_text_on_image_pil(None, text, org)
        import kwplot
        kwplot.autompl()
        kwplot.imshow(new_img)


    """
    from PIL import Image, ImageDraw, ImageFont
    # fontpath = "./simsun.ttc"
    # unicode_font = ImageFont.truetype("DejaVuSans.ttf", font_size)
    if fontpath is None:
        fontpath = "DejaVuSans.ttf"
    # fontpath = "arial.ttf"
    # fontpath = '/home/joncrall/Downloads/Arial.ttf'
    font = ImageFont.truetype(fontpath, fontsize)

    if org is None:
        org = (1, 1)

    if img is None:
        dummy_img = np.empty((8, 8, 4), dtype=np.uint8)
        dummy_img_pil = Image.fromarray(dummy_img)
        dummy_draw = ImageDraw.Draw(dummy_img_pil)
        bbox = dummy_draw.textbbox(org, text, font=font)
        rb_x = bbox[2]
        rb_y = bbox[3]
        img = np.full((rb_y + 1, rb_x + 1, 4), fill_value=0, dtype=np.uint8)
    # size = draw.textlength(text, font=font)
    # print(f'bbox={bbox}')
    # print(f'size={size}')

    img_pil = Image.fromarray(img)
    draw = ImageDraw.Draw(img_pil)
    from kwimage import Color
    color = Color.coerce('black').as255()

    # anchor - https://pillow.readthedocs.io/en/stable/handbook/text-anchors.html#text-anchors
    draw.text(org, text, font=font, fill=color)
    new_img = np.array(img_pil)
    return new_img


def draw_text_on_image(img, text, org=None, return_info=False, **kwargs):
    r"""
    Draws multiline text on an image using opencv

    Args:
        img (ndarray | None | dict):
            Generally a numpy image to draw on (inplace).
            Otherwise a canvas will be constructed such that the text will fit.
            The user may specify a dictionary with keys width and height
            to have more control over the constructed canvas.

        text (str): text to draw

        org (Tuple[int, int]):
            The x, y location of the text string "anchor" in the image as
            specified by halign and valign.  For instance, If valign='bottom',
            halign='left', this where the bottom left corner of the text will
            be placed.

        return_info (bool):
            if True, also returns information about the positions the text
            was drawn on.

        **kwargs:
            color (tuple): default blue

            thickness (int): defaults to 2

            fontFace (int): defaults to cv2.FONT_HERSHEY_SIMPLEX

            fontScale (float): defaults to 1.0

            valign (str):
            either top, center, or bottom.
            Defaults to "bottom"
            NOTE: this default may change to "top" in the future.

            halign (str):
            either left, center, or right. Defaults to "left".

            border (dict | int):
            If specified as an integer, draws a black border with that
            given thickness.  If specified as a dictionary, draws a border
            with color specified parameters.
            "color": border color, defaults to "black".
            "thickness": border thickness, defaults to 1.

    Returns:
        ndarray | Tuple[ndarray, dict] :
            The image that was drawn on and optionally an information
            dictionary if return_info was True.

    Note:
        The image is modified inplace. If the image is non-contiguous then this
        returns a UMat instead of a ndarray, so be careful with that.

    Related:
        The logic in this function is related to the following stack overflow
        posts [SO27647424]_ [SO51285616]_

    References:
        .. [SO27647424] https://stackoverflow.com/questions/27647424/
        .. [SO51285616] https://stackoverflow.com/questions/51285616/opencvs-gettextsize-and-puttext-return-wrong-size-and-chop-letters-with-low

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> import kwimage
        >>> img = kwimage.grab_test_image(space='rgb')
        >>> img2 = kwimage.draw_text_on_image(img.copy(), 'FOOBAR', org=(0, 0), valign='top')
        >>> assert img2.shape == img.shape
        >>> assert np.any(img2 != img)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(img2)
        >>> kwplot.show_if_requested()

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> import kwimage
        >>> # Test valign
        >>> img = kwimage.grab_test_image(space='rgb', dsize=(640, 640))
        >>> img2 = kwimage.draw_text_on_image(img, 'VALIGN-top\nbazbiz\nspam', org=(0, 100), valign='top', border=2)
        >>> img2 = kwimage.draw_text_on_image(img, 'VALIGN-center\nbazbiz\nspam', org=(200, 100), valign='center', border=2)
        >>> img2 = kwimage.draw_text_on_image(img, 'VALIGN-bottom\nbazbiz\nspam', org=(450, 100), valign='bottom', border=2)
        >>> # Test halign
        >>> img2 = kwimage.draw_text_on_image(img, 'HALIGN-right\nbazbiz\nspam', org=(250, 300), halign='right', border=2)
        >>> img2 = kwimage.draw_text_on_image(img, 'HALIGN-center\nbazbiz\nspam', org=(250, 400), halign='center', border=2)
        >>> img2 = kwimage.draw_text_on_image(img, 'HALIGN-left\nbazbiz\nspam', org=(250, 500), halign='left', border=2)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(img2)
        >>> kwplot.show_if_requested()

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> # Ensure the function works with float01 or uint255 images
        >>> import kwimage
        >>> img = kwimage.grab_test_image(space='rgb')
        >>> img = kwimage.ensure_float01(img)
        >>> img2 = img.copy()
        >>> img2 = kwimage.draw_text_on_image(img2, 'FOOBAR\nbazbiz\nspam', org=(0, 0), valign='top', border=2, fontScale=1.0)
        >>> img2 = kwimage.draw_text_on_image(img2, 'FOOBAR\nbazbiz\nspam', org=(0, 200), valign='top', border=2, fontScale=2.0)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(img2)
        >>> kwplot.show_if_requested()

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> # Test dictionary border
        >>> import kwimage
        >>> img = kwimage.draw_text_on_image(None, 'Battery\nFraction', org=(100, 100), valign='top', halign='center', border={'color': 'green', 'thickness': 9})
        >>> #img = kwimage.draw_text_on_image(None, 'hello\neveryone', org=(0, 0), valign='top')
        >>> #img = kwimage.draw_text_on_image(None, 'hello', org=(0, 60), valign='top', halign='center', border=0)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(img)
        >>> kwplot.show_if_requested()

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> # Test dictionary image
        >>> import kwimage
        >>> img = kwimage.draw_text_on_image({'width': 300}, 'Arbitrary\nText', org=(150, 0), valign='top', halign='center', border={'color': 'green', 'thickness': 0})
        >>> print('img.shape = {!r}'.format(img.shape))
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(img)
        >>> kwplot.show_if_requested()

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> # Test fontScale
        >>> import kwimage
        >>> canvases = []
        >>> canvases.append(kwimage.draw_text_on_image(None, 'FontScale=1.0', fontScale=1.0))
        >>> canvases.append(kwimage.draw_text_on_image(None, 'FontScale=2.0', fontScale=2.0))
        >>> canvases.append(kwimage.draw_text_on_image(None, 'FontScale=3.0', fontScale=3.0))
        >>> # xdoctest: +REQUIRES(--show)
        >>> canvas = kwimage.stack_images_grid(canvases, pad=10, bg_value=(255, 255, 255))
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(canvas)
        >>> kwplot.show_if_requested()

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> import ubelt as ub
        >>> import kwimage
        >>> grid = list(ub.named_product({
        >>>     'halign': ['left', 'center', 'right', None],
        >>>     'valign': ['top', 'center', 'bottom', None],
        >>>     'border': [0, 3]
        >>> }))
        >>> canvases = []
        >>> text = 'small-line\na-much-much-much-bigger-line\nanother-small\n.'
        >>> for kw in grid:
        >>>     header = kwimage.draw_text_on_image({}, ub.urepr(kw, compact=1), color='blue')
        >>>     canvas = kwimage.draw_text_on_image({'color': 'white'}, text, org=None, **kw)
        >>>     canvases.append(kwimage.stack_images([header, canvas], axis=0, bg_value=(255, 255, 255), pad=5))
        >>> # xdoctest: +REQUIRES(--show)
        >>> canvas = kwimage.stack_images_grid(canvases, pad=10, bg_value=(255, 255, 255))
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(canvas)
        >>> kwplot.show_if_requested()
    """
    import kwimage
    import cv2

    if 'color' not in kwargs:
        # kwargs['color'] = 'red'
        kwargs['color'] = 'strawberry'

    # Get the color that is compatible with the input image encoding
    if img is None or isinstance(img, dict):
        kwargs['color'] = kwimage.Color(kwargs['color']).as255()
    else:
        kwargs['color'] = kwimage.Color(kwargs['color']).forimage(img)

    if 'thickness' not in kwargs:
        kwargs['thickness'] = 2

    if 'fontFace' not in kwargs:
        kwargs['fontFace'] = cv2.FONT_HERSHEY_SIMPLEX

    if 'fontScale' not in kwargs:
        kwargs['fontScale'] = 1.0

    if 'lineType' not in kwargs:
        kwargs['lineType'] = cv2.LINE_AA

    if 'bottomLeftOrigin' in kwargs:
        raise ValueError('Do not use bottomLeftOrigin, use valign instead')

    border = kwargs.pop('border', None)
    if border is not None:
        if isinstance(border, int):
            border = {'color': 'black', 'thickness': border}
        subkw = kwargs.copy()
        subkw['color'] = border.get('color', 'black')
        subkw.pop('return_info', None)
        border_thickness = border.get('thickness', 1)
    else:
        border_thickness = 0

    valign = kwargs.pop('valign', None)
    halign = kwargs.pop('halign', None)
    if halign is None:
        halign = 'left'
    if valign is None:
        valign = 'top'

    if img is None:
        img = {'width': None, 'height': None}

    if org is None:
        org = (None, None)

    x0, y0 = org

    if isinstance(img, dict):
        given_w = img.get('width', None)
        given_h = img.get('height', None)
    else:
        given_h, given_w = img.shape[0:2]

    needs_x0 = x0 is None and halign != 'left'
    needs_y0 = y0 is None and valign != 'top'

    if needs_x0 or needs_y0:
        # Special case: when the alignment is non left-top, AND we don't have
        # an origin we need to do a bit of extra computation to figure out what
        # the width / height need to be
        text_w, text_h = _text_sizes(text, (1, 1), border_thickness, kwargs, None, halign='left')[0:2]
        if given_w is None:
            given_w = text_w
        if given_h is None:
            given_h = text_h

    if x0 is None:
        if halign == 'left':
            x0 = 1
        elif halign == 'center':
            x0 = given_w // 2
        elif halign == 'right':
            x0 = given_w - 1
        else:
            raise KeyError(halign)

    if y0 is None:
        if valign == 'top':
            y0 = 1
        elif valign == 'center':
            y0 = given_h // 2
        elif valign == 'bottom':
            y0 = given_h - 1
        else:
            raise KeyError(valign)

    org = (x0, y0)
    text_w, text_h, x0, lines, abs_top_y, first_h, total_h, total_w, final_baseline, line_sizes, line_org = _text_sizes(text, org, border_thickness, kwargs, valign, halign)

    if isinstance(img, dict):
        # if image is unspecified allocate just enough space for text
        # allow users to specify partial parameters
        bg_color = kwimage.Color(img.get('color', (0, 0, 0))).as255()
        alloc_w = given_w
        alloc_h = given_h
        if alloc_w is None:
            alloc_w = text_w
        if alloc_h is None:
            alloc_h = text_h
        img = np.zeros((alloc_h, alloc_w, len(bg_color)), dtype=np.uint8)
        img[...] = np.array(bg_color)[None, None, :]
        kwargs['color'] = kwimage.Color(kwargs['color']).forimage(img)

    if border_thickness > 0:
        # recursive call
        basis = list(range(-border_thickness, border_thickness + 1))
        org = np.array(org)
        for i, j in it.product(basis, basis):
            if i == 0 and j == 0:
                continue
            img = draw_text_on_image(img, text, org=org + [i, j], **subkw)

    for i, line in enumerate(lines):
        xy = tuple(line_org[i])
        img = cv2.putText(img, line, xy, **kwargs)

    if return_info:
        info = {
            'line_org': line_org,
            'line_sizes': line_sizes,
        }
        return img, info
    else:
        return img


def _text_sizes(text, org, border_thickness, kwargs, valign, halign):
    import cv2
    getsize_kw = {
        k: kwargs[k]
        for k in ['fontFace', 'fontScale', 'thickness']
        if k in kwargs
    }
    x0, y0 = list(map(int, org))
    thickness = kwargs.get('thickness', 2)

    vertical_spacing = 4  # space between vertical lines
    ypad = thickness + vertical_spacing

    lines = text.split('\n')

    line_sizes = []
    final_baseline = 0
    for line in lines:
        # TODO: better handling of baseline
        # https://en.wikipedia.org/wiki/Baseline_(typography)
        (line_width, line_height), baseline = cv2.getTextSize(line, **getsize_kw)
        line_sizes.append((line_width, line_height))
        final_baseline = baseline
    line_sizes = np.array(line_sizes)

    line_org = []
    y = y0
    for w, h in line_sizes:
        next_y = y + (h + ypad)
        line_org.append((x0, y))
        y = next_y
    line_org = np.array(line_org)

    # the absolute top and bottom position of text
    abs_top_y = line_org[0, 1]
    abs_bot_y = (line_org[-1, 1] + line_sizes[-1, 1]) + thickness

    first_h = line_sizes[0, 1]

    total_h = (abs_bot_y - abs_top_y)
    total_w = line_sizes.T[0].max()

    if valign is not None:
        if valign == 'bottom':
            # This is the default for the one-line case
            # in the multiline case we need to subtract the total
            # height of all lines but the first to ensure the last
            # line is on the bottom.
            line_org[:, 1] -= (total_h - first_h)
        elif valign == 'center':
            # Change from bottom to center
            line_org[:, 1] += first_h - total_h // 2
        elif valign == 'top':
            # Because bottom is the default we just need to add height of the
            # first line.
            line_org[:, 1] += first_h
        else:
            raise KeyError(valign)

    if halign == 'left':
        # This is the default case, no modification needed
        pass
    elif halign == 'center':
        # When the x-orgin should be the center, subtract half of
        # the line width to get the leftmost point.
        line_org[:, 0] = x0 - (line_sizes[:, 0] / 2)
    elif halign == 'right':
        # The x-orgin should be the rightmost point, subtract
        # the width of each line to find the leftmost point.
        line_org[:, 0] = x0 - line_sizes[:, 0]
    else:
        raise KeyError(halign)

    abs_left_x = line_org[:, 0].min()
    text_w = total_w + border_thickness + abs_left_x
    text_h = total_h + border_thickness + abs_top_y + final_baseline

    return text_w, text_h, x0, lines, abs_top_y, first_h, total_h, total_w, final_baseline, line_sizes, line_org


def draw_clf_on_image(im, classes, tcx=None, probs=None, pcx=None, border=1):
    """
    Draws classification label on an image.

    Works best with image chips sized between 200x200 and 500x500

    Args:
        im (ndarray): the image
        classes (Sequence[str] | kwcoco.CategoryTree): list of class names
        tcx (int): true class index if known
        probs (ndarray): predicted class probs for each class
        pcx (int): predicted class index.
            (if None but probs is specified uses argmax of probs)

    Example:
        >>> # xdoctest: +REQUIRES(module:torch)
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> import torch
        >>> import kwarray
        >>> import kwimage
        >>> rng = kwarray.ensure_rng(0)
        >>> im = (rng.rand(300, 300) * 255).astype(np.uint8)
        >>> classes = ['cls_a', 'cls_b', 'cls_c']
        >>> tcx = 1
        >>> probs = rng.rand(len(classes))
        >>> probs[tcx] = 0
        >>> probs = torch.FloatTensor(probs).softmax(dim=0).numpy()
        >>> im1_ = kwimage.draw_clf_on_image(im, classes, tcx, probs)
        >>> probs[tcx] = .9
        >>> probs = torch.FloatTensor(probs).softmax(dim=0).numpy()
        >>> im2_ = kwimage.draw_clf_on_image(im, classes, tcx, probs)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(im1_, colorspace='rgb', pnum=(1, 2, 1), fnum=1, doclf=True)
        >>> kwplot.imshow(im2_, colorspace='rgb', pnum=(1, 2, 2), fnum=1)
        >>> kwplot.show_if_requested()
    """
    import kwimage
    im_ = kwimage.atleast_3channels(im)
    w, h = im.shape[0:2][::-1]

    if pcx is None and probs is not None:
        import kwarray
        probs = kwarray.ArrayAPI.numpy(probs)
        pcx = probs.argmax()

    if probs is not None:
        pred_score = None if pcx is None else probs[pcx]
        true_score = None if tcx is None else probs[tcx]

    org1 = np.array((2, h - 5))
    org2 = np.array((2, 5))

    true_label = None
    if tcx is not None:
        true_name = classes[tcx]
        if pcx == tcx:
            true_label = 't:{tcx}:{true_name}'.format(**locals())
        elif probs is None:
            true_label = 't:{tcx}:\n{true_name}'.format(**locals())
        else:
            true_label = 't:{tcx}@{true_score:.2f}:\n{true_name}'.format(**locals())

    pred_label = None
    if pcx is not None:
        pred_name = classes[pcx]
        if probs is None:
            pred_label = 'p:{pcx}:\n{pred_name}'.format(**locals())
        else:
            pred_label = 'p:{pcx}@{pred_score:.2f}:\n{pred_name}'.format(**locals())

    fontkw = {
        'fontScale': 1.0,
        'thickness': 2
    }
    color = 'dodgerblue' if pcx == tcx else 'orangered'

    # im_ = draw_text_on_image(im_, pred_label, org=org1 - 2,
    #                          color='white', valign='bottom', **fontkw)
    # im_ = draw_text_on_image(im_, true_label, org=org2 - 2,
    #                          color='white', valign='top', **fontkw)

    if pred_label is not None:
        im_ = draw_text_on_image(im_, pred_label, org=org1, color=color,
                                 border=border, valign='bottom', **fontkw)
    if true_label is not None:
        im_ = draw_text_on_image(im_, true_label, org=org2, color='lawngreen',
                                 valign='top', border=border, **fontkw)
    return im_


def draw_boxes_on_image(img, boxes, color='blue', thickness=1,
                        box_format=None, colorspace='rgb'):
    """
    Draws boxes on an image.

    Args:
        img (ndarray): image to copy and draw on
        boxes (kwimage.Boxes | ndarray): boxes to draw
        colorspace (str): string code of the input image colorspace

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> import kwimage
        >>> import numpy as np
        >>> img = np.zeros((10, 10, 3), dtype=np.uint8)
        >>> color = 'dodgerblue'
        >>> thickness = 1
        >>> boxes = kwimage.Boxes([[1, 1, 8, 8]], 'ltrb')
        >>> img2 = draw_boxes_on_image(img, boxes, color, thickness)
        >>> assert tuple(img2[1, 1]) == (30, 144, 255)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()  # xdoctest: +SKIP
        >>> kwplot.figure(doclf=True, fnum=1)
        >>> kwplot.imshow(img2)
    """
    import kwimage
    import cv2
    if not isinstance(boxes, kwimage.Boxes):
        if box_format is None:
            raise ValueError('specify box_format')
        boxes = kwimage.Boxes(boxes, box_format)

    color = kwimage.Color(color).forimage(img, colorspace)
    ltrb = boxes.to_ltrb().data
    img2 = img.copy()
    for x1, y1, x2, y2 in ltrb:
        # pt1 = (int(round(x1)), int(round(y1)))
        # pt2 = (int(round(x2)), int(round(y2)))
        pt1 = (int(x1), int(y1))
        pt2 = (int(x2), int(y2))
        # Note cv2.rectangle does work inplace
        img2 = cv2.rectangle(img2, pt1, pt2, color, thickness=thickness)
    return img2


def draw_line_segments_on_image(
        img, pts1, pts2, color='blue', colorspace='rgb', thickness=1,
        **kwargs):
    """
    Draw line segments between pts1 and pts2 on an image.

    Args:
        img (ndarray): image to draw on
        pts1 (ndarray): xy coordinates of starting points
        pts2 (ndarray): corresponding xy coordinates of ending points
        color (str | List):
            color code or a list of colors for each line segment
        colorspace (str): colorspace of image. Defaults to 'rgb'
        thickness (int): Defaults to 1
        lineType (int): option for cv2.line

    Returns:
        ndarray: the modified image (inplace if possible)

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> from kwimage.im_draw import *  # NOQA
        >>> pts1 = np.array([[2, 0], [2, 20], [2.5, 30]])
        >>> pts2 = np.array([[10, 5], [30, 28], [100, 50]])
        >>> img = np.ones((100, 100, 3), dtype=np.uint8) * 255
        >>> color = 'blue'
        >>> colorspace = 'rgb'
        >>> img2 = draw_line_segments_on_image(img, pts1, pts2, thickness=2)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()  # xdoctest: +SKIP
        >>> kwplot.figure(doclf=True, fnum=1)
        >>> kwplot.imshow(img2)

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> # xdoctest: +REQUIRES(module:matplotlib)
        >>> import kwimage
        >>> pts1 = kwimage.Points.random(10).scale(512).xy
        >>> pts2 = kwimage.Points.random(10).scale(512).xy
        >>> img = np.ones((512, 512, 3), dtype=np.uint8) * 255
        >>> color = kwimage.Color.distinct(10)
        >>> img2 = kwimage.draw_line_segments_on_image(img, pts1, pts2, color=color)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()  # xdoctest: +SKIP
        >>> kwplot.figure(doclf=True, fnum=1)
        >>> kwplot.imshow(img2)
    """
    import cv2
    num = len(pts1)
    colors = _broadcast_colors(color, num, img, colorspace)

    if 'lineType' not in kwargs:
        kwargs['lineType'] = cv2.LINE_AA

    pts1_ = pts1.tolist()
    pts2_ = pts2.tolist()
    for xy1, xy2, col in zip(pts1_, pts2_, colors):
        xy1 = tuple(map(int, xy1))
        xy2 = tuple(map(int, xy2))
        cv2.line(img, xy1, xy2, color=col, thickness=thickness, **kwargs)
    return img


def draw_polyline_on_image(image, xy_pts, edgecolor='blue', thickness=1, **kwargs):
    """
    Draw a path (i.e. polyline / linestring) on an image.

    Args:
        img (ndarray): image to draw on
        xy_pts (ndarray): sequence of ordered xy coordinates in a polyline
        edgecolor (Color | list[Color]):
            a single color or a list of colors for each edge between points
            (i.e. has length len(xy_pts) - 1)

    Returns:
        ndarray: the modified image (inplace if possible)

    Example:
        >>> from kwimage.im_draw import *  # NOQA
        >>> import kwimage
        >>> pts = kwimage.Points.random(10).scale(512)
        >>> xy_pts = pts.xy
        >>> image = kwimage.checkerboard(dsize=(512, 512))
        >>> image = kwimage.atleast_3channels(image)
        >>> image = kwimage.ensure_uint255(image)
        >>> edgecolor = [kwimage.Color.random().as255() for _ in range(len(xy_pts) - 1)]
        >>> thickness = 5
        >>> image = draw_polyline_on_image(image, xy_pts, edgecolor=edgecolor, thickness=thickness)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(image)
    """
    pts1 = xy_pts[0:-1]
    pts2 = xy_pts[1:]
    if len(pts1):
        # Draw edges.
        image = draw_line_segments_on_image(
            image, pts1, pts2, color=edgecolor, thickness=thickness, **kwargs)
    return image


def _broadcast_colors(color, num, img, colorspace):
    """
    Determine if color applies a single color to all ``num`` items, or if it is
    a list of colors for each item. Return as a list of colors for each item.

    TODO:
        - [ ] add as classmethod of kwimage.Color

    Example:
        >>> img = (np.random.rand(512, 512, 3) * 255).astype(np.uint8)
        >>> colorspace = 'rgb'
        >>> color = color_str_list = ['red', 'green', 'blue']
        >>> color_str = 'red'
        >>> num = 3
        >>> print(_broadcast_colors(color_str_list, num, img, colorspace))
        >>> print(_broadcast_colors(color_str, num, img, colorspace))
        >>> colors_tuple_list = _broadcast_colors(color_str_list, num, img, colorspace)
        >>> print(_broadcast_colors(colors_tuple_list, num, img, colorspace))
        >>> #
        >>> # FIXME: This case seems broken
        >>> colors_ndarray_list = np.array(_broadcast_colors(color_str_list, num, img, colorspace))
        >>> print(_broadcast_colors(colors_ndarray_list, num, img, colorspace))
    """
    # Note there is an ambiguity when num=3 and color=[int, int, int]
    # that must be resolved by checking num channels in the image
    import kwimage
    import ubelt as ub
    import numbers

    needs_broadcast = True  # assume the list wasnt given by default
    if ub.iterable(color):
        first = ub.peek(color)
        if len(color) == num:
            if len(color) <= 4 and isinstance(first, numbers.Number):
                # ambiguous case, interpret as a single broadcastable color
                needs_broadcast = True
            else:
                # This is the only case we dont need broadcast
                needs_broadcast = False

    if needs_broadcast:
        color = kwimage.Color(color).forimage(img, colorspace)
        colors = [color] * num
    else:
        colors = [kwimage.Color(c).forimage(img, colorspace) for c in color]
    return colors


def make_heatmask(probs, cmap='plasma', with_alpha=1.0, space='rgb',
                  dsize=None):
    """
    Colorizes a single-channel intensity mask (with an alpha channel)

    Args:
        probs (ndarray): 2D probability map with values between 0 and 1
        cmap (str): mpl colormap
        with_alpha (float): between 0 and 1, uses probs as the alpha multipled
            by this number.
        space (str): output colorspace
        dsize (tuple): if not None, then output is resized to W,H=dsize

    SeeAlso:
        kwimage.overlay_alpha_images

    Example:
        >>> # xdoctest: +REQUIRES(module:matplotlib)
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> from kwimage.im_draw import *  # NOQA
        >>> probs = np.tile(np.linspace(0, 1, 10), (10, 1))
        >>> heatmask = make_heatmask(probs, with_alpha=0.8, dsize=(100, 100))
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(heatmask, fnum=1, doclf=True, colorspace='rgb',
        >>>               title='make_heatmask')
        >>> kwplot.show_if_requested()
    """
    import kwimage
    import matplotlib as mpl
    import matplotlib.cm  # NOQA
    assert len(probs.shape) == 2
    try:
        cmap_ = mpl.colormaps[cmap]
    except Exception:
        cmap_ = mpl.cm.get_cmap(cmap)
    probs = kwimage.ensure_float01(probs)
    heatmask = cmap_(probs).astype(np.float32)
    heatmask = kwimage.convert_colorspace(heatmask, 'rgba', space, implicit=True)
    if with_alpha is not False and with_alpha is not None:
        heatmask[:, :, 3] = (probs * with_alpha)  # assign probs to alpha channel
    if dsize is not None:
        import cv2
        heatmask = cv2.resize(
            heatmask, tuple(dsize),
            interpolation=cv2.INTER_NEAREST)
    return heatmask


def make_orimask(radians, mag=None, alpha=1.0):
    """
    Makes a colormap in HSV space where the orientation changes color and mag
    changes the saturation/value.

    Args:
        radians (ndarray): orientation in radians
        mag (ndarray): magnitude (must be normalized between 0 and 1)
        alpha (float | ndarray):
            if False or None, then the image is returned without alpha
            if a float, then mag is scaled by this and used as the alpha channel
            if an ndarray, then this is explicilty set as the alpha channel

    Returns:
        ndarray[Any, Float32]:
            an rgb / rgba image in 01 space

    SeeAlso:
        kwimage.overlay_alpha_images

    Example:
        >>> # xdoctest: +REQUIRES(module:matplotlib)
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> from kwimage.im_draw import *  # NOQA
        >>> x, y = np.meshgrid(np.arange(64), np.arange(64))
        >>> dx, dy = x - 32, y - 32
        >>> radians = np.arctan2(dx, dy)
        >>> mag = np.sqrt(dx ** 2 + dy ** 2)
        >>> orimask = make_orimask(radians, mag)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(orimask, fnum=1, doclf=True,
        >>>               colorspace='rgb', title='make_orimask')
        >>> kwplot.show_if_requested()
    """
    import matplotlib as mpl
    import matplotlib.cm  # NOQA
    TAU = np.pi * 2
    # Map radians to 0 to 1
    ori01 = (radians % TAU) / TAU
    try:
        cmap_ = mpl.colormaps['hsv']
    except Exception:
        cmap_ = mpl.cm.get_cmap('hsv')
    color_rgb = cmap_(ori01)[..., 0:3].astype(np.float32)
    if mag is not None:
        import kwimage
        if mag.max() > 1:
            mag = mag / mag.max()
        color_hsv = kwimage.convert_colorspace(color_rgb, 'rgb', 'hsv')
        color_hsv[..., 1:3] = mag[..., None]
        color_rgb = kwimage.convert_colorspace(color_hsv, 'hsv', 'rgb')
    else:
        mag = 1
    orimask = np.array(color_rgb, dtype=np.float32)

    if isinstance(alpha, np.ndarray):
        # Alpha specified as explicit numpy array
        orimask = kwimage.ensure_alpha_channel(orimask)
        orimask[:, :, 3] = alpha
    elif alpha is not False and alpha is not None:
        orimask = kwimage.ensure_alpha_channel(orimask)
        orimask[:, :, 3] = mag * alpha
    return orimask


def make_vector_field(dx, dy, stride=0.02, thresh=0.0, scale=1.0, alpha=1.0,
                      color='strawberry', thickness=1, tipLength=0.1, line_type='aa'):
    """
    Create an image representing a 2D vector field.

    Args:
        dx (ndarray): grid of vector x components

        dy (ndarray): grid of vector y components

        stride (int | float): sparsity of vectors, int specifies stride step in
            pixels, a float specifies it as a percentage.

        thresh (float): only plot vectors with magnitude greater than thres

        scale (float): multiply magnitude for easier visualization

        alpha (float): alpha value for vectors. Non-vector regions receive 0
            alpha (if False, no alpha channel is used)

        color (str | tuple | kwimage.Color): RGB color of the vectors

        thickness (int): thickness of arrows

        tipLength (float): fraction of line length

        line_type (int | str):
            either cv2.LINE_4, cv2.LINE_8, or cv2.LINE_AA or a string
            code.

    Returns:
        ndarray[Any, Float32]:
            vec_img - an rgb/rgba image in 0-1 space

    SeeAlso:
        kwimage.overlay_alpha_images

    DEPRECATED USE: draw_vector_field instead

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> x, y = np.meshgrid(np.arange(512), np.arange(512))
        >>> dx, dy = x - 256.01, y - 256.01
        >>> radians = np.arctan2(dx, dy)
        >>> mag = np.sqrt(dx ** 2 + dy ** 2)
        >>> dx, dy = dx / mag, dy / mag
        >>> img = make_vector_field(dx, dy, scale=10, alpha=False)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(img)
        >>> kwplot.show_if_requested()
    """
    # import warnings
    import ubelt as ub
    ub.schedule_deprecation('kwimage', 'make_vector_field', 'method',
                            migration='use draw_vector_field instead',
                            deprecate='0.9.1', error='1.0.0', remove='1.1.0')
    import cv2
    import kwimage
    color = kwimage.Color(color).as255('rgb')
    vecmask = np.zeros(dx.shape + (3,), dtype=np.uint8)

    line_type_lookup = {'aa': cv2.LINE_AA}
    line_type = line_type_lookup.get(line_type, line_type)

    width = dx.shape[1]
    height = dy.shape[0]

    x_grid = np.arange(0, width, 1)
    y_grid = np.arange(0, height, 1)
    # Vector locations and directions
    X, Y = np.meshgrid(x_grid, y_grid)
    U, V = dx, dy

    XYUV = [X, Y, U, V]

    if isinstance(stride, float):
        if stride < 0 or stride > 1:
            raise ValueError('Floating point strides must be between 0 and 1')
        stride = int(np.ceil(stride * min(width, height)))

    # stride the points
    if stride is not None and stride > 1:
        XYUV = [a[::stride, ::stride] for a in XYUV]

    # flatten the points
    XYUV = [a.ravel() for a in XYUV]

    # Filter out points with low magnitudes
    if thresh is not None and thresh > 0:
        M = np.sqrt((XYUV[2] ** 2) + (XYUV[3] ** 2)).ravel()
        XYUV = np.array(XYUV)
        flags = M > thresh
        XYUV = [a[flags] for a in XYUV]

    # Adjust vector magnitude for visibility
    if scale is not None:
        XYUV[2] *= scale
        XYUV[3] *= scale

    for (x, y, u, v) in zip(*XYUV):
        pt1 = (int(x), int(y))
        pt2 = tuple(map(int, map(np.round, (x + u, y + v))))
        cv2.arrowedLine(vecmask, pt1, pt2, color=color, thickness=thickness,
                        tipLength=tipLength,
                        line_type=line_type)

    vecmask = kwimage.ensure_float01(vecmask)
    if isinstance(alpha, np.ndarray):
        # Alpha specified as explicit numpy array
        vecmask = kwimage.ensure_alpha_channel(vecmask)
        vecmask[:, :, 3] = alpha
    elif alpha is not False and alpha is not None:
        # Alpha specified as a scale factor
        vecmask = kwimage.ensure_alpha_channel(vecmask)
        # vecmask[:, :, 3] = (vecmask[:, :, 0:3].sum(axis=2) > 0) * alpha
        vecmask[:, :, 3] = vecmask[:, :, 0:3].sum(axis=2) * alpha
    return vecmask


def draw_vector_field(image, dx, dy, stride=0.02, thresh=0.0, scale=1.0,
                      alpha=1.0, color='strawberry', thickness=1, tipLength=0.1,
                      line_type='aa'):
    """
    Create an image representing a 2D vector field.

    Args:
        image (ndarray): image to draw on

        dx (ndarray): grid of vector x components

        dy (ndarray): grid of vector y components

        stride (int | float): sparsity of vectors, int specifies stride step in
            pixels, a float specifies it as a percentage.

        thresh (float): only plot vectors with magnitude greater than thres

        scale (float): multiply magnitude for easier visualization

        alpha (float): alpha value for vectors. Non-vector regions receive 0
            alpha (if False, no alpha channel is used)

        color (str | tuple | kwimage.Color): RGB color of the vectors

        thickness (int): thickness of arrows

        tipLength (float): fraction of line length

        line_type (int | str):
            either cv2.LINE_4, cv2.LINE_8, or cv2.LINE_AA or 'aa'

    Returns:
        ndarray[Any, Float32]:
            The image with vectors overlaid. If image=None, then an rgb/a image
            is created and returned.

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> from kwimage.im_draw import *  # NOQA
        >>> import kwimage
        >>> width, height = 512, 512
        >>> image = kwimage.grab_test_image(dsize=(width, height))
        >>> x, y = np.meshgrid(np.arange(height), np.arange(width))
        >>> dx, dy = x - width / 2, y - height / 2
        >>> radians = np.arctan2(dx, dy)
        >>> mag = np.sqrt(dx ** 2 + dy ** 2) + 1e-3
        >>> dx, dy = dx / mag, dy / mag
        >>> img = kwimage.draw_vector_field(image, dx, dy, scale=10, alpha=False)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(img, title='draw_vector_field')
        >>> kwplot.show_if_requested()
    """
    import cv2
    import kwimage

    if image is None:
        # Create a default image
        image = np.zeros(dx.shape + (3,), dtype=np.uint8)
        # image = kwimage.atleast_3channels(image)

    color = kwimage.Color(color).forimage(image)

    line_type_lookup = {'aa': cv2.LINE_AA}
    line_type = line_type_lookup.get(line_type, line_type)

    height, width = dx.shape[0:2]

    x_grid = np.arange(0, width, 1)
    y_grid = np.arange(0, height, 1)
    # Vector locations and directions
    X, Y = np.meshgrid(x_grid, y_grid)
    U, V = dx, dy

    XYUV = [X, Y, U, V]

    if isinstance(stride, float):
        if stride < 0 or stride > 1:
            raise ValueError('Floating point strides must be between 0 and 1')
        stride = int(np.ceil(stride * min(width, height)))

    # stride the points
    if stride is not None and stride > 1:
        XYUV = [a[::stride, ::stride] for a in XYUV]

    # flatten the points
    XYUV = [a.ravel() for a in XYUV]

    # Filter out points with low magnitudes
    if thresh is not None and thresh > 0:
        M = np.sqrt((XYUV[2] ** 2) + (XYUV[3] ** 2)).ravel()
        XYUV = np.array(XYUV)
        flags = M > thresh
        XYUV = [a[flags] for a in XYUV]

    # Adjust vector magnitude for visibility
    if scale is not None:
        XYUV[2] *= scale
        XYUV[3] *= scale

    if alpha is not None and alpha is not False and alpha != 1:
        raise NotImplementedError

    for (x, y, u, v) in zip(*XYUV):
        pt1 = (int(x), int(y))
        pt2 = tuple(map(int, map(np.round, (x + u, y + v))))
        cv2.arrowedLine(image, pt1, pt2, color=color, thickness=thickness,
                        tipLength=tipLength,
                        line_type=line_type)

    if isinstance(alpha, np.ndarray):
        # Alpha specified as explicit numpy array
        image = kwimage.ensure_float01(image)
        image = kwimage.ensure_alpha_channel(image)
        image[:, :, 3] = alpha
    elif alpha is not False and alpha is not None:
        # Alpha specified as a scale factor
        image = kwimage.ensure_float01(image)
        image = kwimage.ensure_alpha_channel(image)
        # image[:, :, 3] = (image[:, :, 0:3].sum(axis=2) > 0) * alpha
        image[:, :, 3] = image[:, :, 0:3].sum(axis=2) * alpha
    return image


def draw_header_text(image=None, text=None, fit=False, color='strawberry',
                     halign='center', stack='auto', bg_color='black',
                     **kwargs):
    """
    Places a black bar on top of an image and writes text in it

    Args:

        image (ndarray | dict | None):
            numpy image or dictionary containing a key width

        text (str) :
            text to draw

        fit (bool | str):
            If False, will draw as much text within the given width as possible.
            If True, will draw all text and then resize to fit in the given width
            If "shrink", will only resize the text if it is too big to fit, in
            other words this is like fit=True, but it wont enlarge the text.

        color (str | Tuple) :
            a color coercable to :class:`kwimage.Color`.

        halign (str) :
            Horizontal alignment. Can be left, center, or right.

        stack (bool | str):
            if True returns the stacked image, otherwise just returns the
            header. If 'auto', will only stack if an image is given as an
            ndarray.

        **kwargs: used only for parameter aliases.
            Currently accepts:
                * bg_value as an alias for bg_color.
                * fontScale
                * fontFace
                * thickness

    Returns:
        ndarray

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> from kwimage.im_draw import *  # NOQA
        >>> import kwimage
        >>> image = kwimage.grab_test_image()
        >>> tiny_image = kwimage.imresize(image, dsize=(64, 64))
        >>> canvases = []
        >>> canvases += [draw_header_text(image=image, text='unfit long header ' * 5, fit=False)]
        >>> canvases += [draw_header_text(image=image, text='shrunk long header ' * 5, fit='shrink')]
        >>> canvases += [draw_header_text(image=image, text='left header', fit=False, halign='left')]
        >>> canvases += [draw_header_text(image=image, text='center header', fit=False, halign='center')]
        >>> canvases += [draw_header_text(image=image, text='right header', fit=False, halign='right')]
        >>> canvases += [draw_header_text(image=image, text='shrunk header', fit='shrink', halign='left')]
        >>> canvases += [draw_header_text(image=tiny_image, text='shrunk header-center', fit='shrink', halign='center')]
        >>> canvases += [draw_header_text(image=image, text='fit header', fit=True, halign='left')]
        >>> canvases += [draw_header_text(image={'width': 200}, text='header only', fit=True, halign='left')]
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> pnum_ = kwplot.PlotNums(nCols=3, nSubplots=len(canvases))
        >>> for c in canvases:
        >>>     kwplot.imshow(c, pnum=pnum_())
        >>> kwplot.show_if_requested()

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> # Test fontScale
        >>> from kwimage.im_draw import *  # NOQA
        >>> import kwimage
        >>> test_image = kwimage.grab_test_image(dsize=(900, 120))
        >>> canvases = []
        >>> for image in [test_image, None]:
        >>>     canvases += [draw_header_text(image=image, text='Scale=1.0 Fit=False', fit=False, fontScale=1.0)]
        >>>     canvases += [draw_header_text(image=image, text='Scale=2.0 Fit=False', fit=False, fontScale=2.0)]
        >>>     canvases += [draw_header_text(image=image, text='Scale=6.0 Fit=False', fit=False, fontScale=6.0)]
        >>>     canvases += [draw_header_text(image=image, text='Scale=1.0 Fit=True', fit=True, fontScale=1.0)]
        >>>     canvases += [draw_header_text(image=image, text='Scale=2.0 Fit=True', fit=True, fontScale=2.0)]
        >>>     canvases += [draw_header_text(image=image, text='Scale=6.0 Fit=True', fit=True, fontScale=6.0)]
        >>>     canvases += [draw_header_text(image=image, text='Scale=1.0 Fit=shrink', fit='shrink', fontScale=1.0)]
        >>>     canvases += [draw_header_text(image=image, text='Scale=2.0 Fit=shrink', fit='shrink', fontScale=2.0)]
        >>>     canvases += [draw_header_text(image=image, text='Scale=6.0 Fit=shrink', fit='shrink', fontScale=6.0)]
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> pnum_ = kwplot.PlotNums(nCols=3, nSubplots=len(canvases))
        >>> for c in canvases:
        >>>     kwplot.imshow(c, pnum=pnum_(), title=str(c.shape))
        >>> kwplot.show_if_requested()
    """
    # import cv2
    import kwimage
    import ubelt as ub

    if text is None:
        raise ValueError('text must be provided')

    if stack == 'auto':
        stack = isinstance(image, np.ndarray)

    if image is None:
        width = None
    elif isinstance(image, dict):
        width = image['width']
        if stack:
            raise ValueError('Must pass in the actual image if stack is True')
    else:
        width = image.shape[1]

    if stack and image is not None:
        # Handle very small image case
        h, w = image.shape[0:2]
        min_pixels = 32
        if w < min_pixels or h < min_pixels:
            image = kwimage.imresize(image, min_dim=min_pixels)
        width = image.shape[1]

    if 'bg_value' in kwargs:
        bg_color = kwargs.pop('bg_value')

    draw_keys = [
        'fontScale',
        'fontFace',
        'thickness',
    ]
    kwargs = ub.udict(kwargs)
    default_draw_kw = ub.udict({
        'valign': 'top',
        'halign': halign,
        'org': None,
        'color': color,
    })
    draw_kw = default_draw_kw | (ub.udict(kwargs) & draw_keys)
    kwargs -= draw_keys

    if kwargs:
        raise ValueError('Unexpected kwargs = {}'.format(kwargs))

    bginfo = {'color': bg_color}

    if fit:
        # TODO: allow a shrink-to-fit only option
        try:
            # needs new kwimage to work
            draw_kw['org'] = None
            header = kwimage.draw_text_on_image(bginfo, text, **draw_kw)
        except Exception:
            draw_kw['halign'] = 'left'
            draw_kw['org'] = (1, 1)
            header = kwimage.draw_text_on_image(bginfo, text, **draw_kw)

        if fit == 'shrink':
            if width is not None:
                if header.shape[1] > width:
                    header = kwimage.imresize(header, dsize=(width, None))
                elif header.shape[1] < width:
                    header = np.pad(header, [(0, 0), ((width - header.shape[1]) // 2, 0), (0, 0)])
                else:
                    pass
        else:
            header = kwimage.imresize(header, dsize=(width, None))
    else:
        # Allows for however much height is needed
        if halign == 'left':
            draw_kw['org'] = (1, 1)
        elif halign == 'center':
            if width is not None:
                draw_kw['org'] = (width // 2, 1)
        elif halign == 'right':
            if width is not None:
                draw_kw['org'] = (width - 1, 1)
        else:
            raise KeyError(halign)

        bginfo['width'] = width
        header = kwimage.draw_text_on_image(bginfo, text, **draw_kw)

    if stack:
        stacked = kwimage.stack_images([header, image], axis=0, overlap=-1)
        return stacked
    else:
        return header


def fill_nans_with_checkers(canvas, square_shape=8,
                            on_value='auto', off_value='auto'):
    """
    Fills nan or masked values with a 2d checkerboard pattern.

    Args:
        canvas (np.ndarray): data replace nans in

        square_shape (int | Tuple[int, int] | str):
            Size of the checker squares. Defaults to 8.

        on_value (Number | str):
            The value of one checker. Defaults to a dark-gray color, 0.3 for
            floats and 77 for ints.

        off_value (Number | str):
            The value off the other checker. Defaults to black, which is 0.

    Returns:
        np.ndarray: the inplace modified canvas

    SeeAlso:
        :func:`nodata_checkerboard` - similar, but operates on nans or masked arrays.

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> from kwimage.im_draw import *  # NOQA
        >>> import kwimage
        >>> orig_img = kwimage.ensure_float01(kwimage.grab_test_image())
        >>> poly1 = kwimage.Polygon.random(rng=1).scale(orig_img.shape[0] // 2)
        >>> poly2 = kwimage.Polygon.random(rng=3).scale(orig_img.shape[0])
        >>> poly3 = kwimage.Polygon.random(rng=4).scale(orig_img.shape[0] // 2)
        >>> poly3 = poly3.translate((0, 200))
        >>> poly4 = poly2.translate((100, 0))
        >>> poly5 = poly2.translate((50, 100))
        >>> img = orig_img.copy()
        >>> img = poly1.fill(img, np.nan)
        >>> img = poly3.fill(img, 0)
        >>> img[:, :, 0] = poly2.fill(np.ascontiguousarray(img[:, :, 0]), np.nan)
        >>> img[:, :, 2] = poly4.fill(np.ascontiguousarray(img[:, :, 2]), np.nan)
        >>> img[:, :, 1] = poly5.fill(np.ascontiguousarray(img[:, :, 1]), np.nan)
        >>> input_img = img.copy()
        >>> canvas = fill_nans_with_checkers(input_img, on_value=0.3)
        >>> assert input_img is canvas
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(img, pnum=(1, 2, 1), title='matplotlib treats nans as zeros')
        >>> kwplot.imshow(canvas, pnum=(1, 2, 2), title='checkers highlight real nans')

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> # Test grayscale
        >>> from kwimage.im_draw import *  # NOQA
        >>> import kwimage
        >>> orig_img = kwimage.ensure_float01(kwimage.grab_test_image())
        >>> poly1 = kwimage.Polygon.random().scale(orig_img.shape[0] // 2)
        >>> poly2 = kwimage.Polygon.random().scale(orig_img.shape[0])
        >>> img = orig_img.copy()
        >>> img = poly1.fill(img, np.nan)
        >>> img[:, :, 0] = poly2.fill(np.ascontiguousarray(img[:, :, 0]), np.nan)
        >>> img = kwimage.convert_colorspace(img, 'rgb', 'gray')
        >>> canvas = img.copy()
        >>> canvas = fill_nans_with_checkers(canvas)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(img, pnum=(1, 2, 1))
        >>> kwplot.imshow(canvas, pnum=(1, 2, 2))

    Ignore:
        >>> import kwarray
        >>> import numpy as np
        >>> img = np.array([[
        >>>     [   0.5,    0.5,    0.5],
        >>>     [np.nan,    0.5,    0.5],
        >>>     [   0.5, np.nan,    0.5],
        >>>     [np.nan, np.nan,    0.5],
        >>>     [   0.5,    0.5, np.nan],
        >>>     [np.nan,    0.5, np.nan],
        >>>     [   0.5, np.nan, np.nan],
        >>>     [np.nan, np.nan, np.nan],
        >>> ]])
        >>> canvas = kwimage.fill_nans_with_checkers(img, square_shape=1)
        >>> print(ub.urepr({'canvas': canvas}, nl=2, with_dtype=False))
        >>> print(canvas)
    """
    invalid_mask = np.isnan(canvas)
    return _masked_checkerboard(canvas, invalid_mask, square_shape, on_value=on_value, off_value=off_value)


def _masked_checkerboard(canvas, invalid_mask, square_shape, on_value, off_value):
    import kwimage
    import kwarray
    canvas = kwarray.atleast_nd(canvas, 3)
    invalid_mask = kwarray.atleast_nd(invalid_mask, 3)
    allchan_invalid_mask = invalid_mask.all(axis=2, keepdims=True)
    anychan_invalid_mask = invalid_mask.any(axis=2, keepdims=True)

    some_invalid_mask = (~allchan_invalid_mask) * anychan_invalid_mask
    dsize = canvas.shape[0:2][::-1]

    if on_value == 'auto':
        if canvas.dtype.kind == 'u' and canvas.dtype.itemsize == 1:
            on_value = 77  # 255
        else:
            on_value = 0.3  # 1
    if off_value == 'auto':
        off_value = 0

    any_total_nans = np.any(allchan_invalid_mask)
    any_partial_nans = np.any(some_invalid_mask)

    if any_total_nans or any_partial_nans:
        checkers2d = kwimage.checkerboard(
            square_shape=square_shape, dsize=dsize, dtype=canvas.dtype,
            on_value=on_value)

        if any_total_nans:
            # canvas = kwimage.ensure_alpha_channel(canvas, (1 - invalid_mask))
            # checkers = kwimage.ensure_alpha_channel(checkers, 1)
            locs = np.where(allchan_invalid_mask)
            canvas[locs[0:2]] = checkers2d[..., None][locs[0:2]]

        if any_partial_nans:
            for chan_idx in range(invalid_mask.shape[2]):
                chan_mask = invalid_mask[..., chan_idx]
                locs3d = np.where(chan_mask[..., None])
                locs3d[2][:] = chan_idx
                locs2d = locs3d[0:2]
                canvas[locs3d] = checkers2d[locs2d]

    return canvas


def nodata_checkerboard(canvas, square_shape=8, on_value='auto', off_value='auto'):
    """
    Fills nans or masked values with a checkerbord pattern.

    Args:
        canvas (ndarray): A 2D image with any number of channels that may be a
            masked array or contain nan values.

        square_shape (int): the pixel size of the checkers

        on_value (Number | str):
            The value of one checker. Defaults to 1 for floats and 255 for
            ints.

        off_value (Number | str):
            The value off the other checker. Defaults to 0.


    Returns:
        ndarray : an output array with imputed values.
            if the input was a masked array, the mask will still exist.

    SeeAlso:
        :func:`fill_nans_with_checkers` - similar, but only operates on nan
        values.

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> import kwimage
        >>> # Test a masked array WITH nan values
        >>> data = kwimage.grab_test_image(space='rgb')
        >>> na_circle = kwimage.Polygon.circle((256 - 96, 256), 128)
        >>> ma_circle = kwimage.Polygon.circle((256 + 96, 256), 128)
        >>> ma_mask = na_circle.fill(np.zeros(data.shape, dtype=np.uint8), value=1).astype(bool)
        >>> na_mask = ma_circle.fill(np.zeros(data.shape, dtype=np.uint8), value=1).astype(bool)
        >>> # Hack the channels to make a ven diagram
        >>> ma_mask[..., [0, 1]] = False
        >>> na_mask[..., [0, 2]] = False
        >>> data = kwimage.ensure_float01(data)
        >>> data[na_mask] = np.nan
        >>> canvas = np.ma.MaskedArray(data, ma_mask)
        >>> kwimage.draw_text_on_image(canvas, 'masked values', (256 - 96, 256 - 128), halign='center', valign='bottom', border=2)
        >>> kwimage.draw_text_on_image(canvas, 'nan values',    (256 + 96, 256 + 128), halign='center', valign='top', border=2)
        >>> kwimage.draw_text_on_image(canvas, 'kwimage.nodata_checkerboard',    (256, 5), halign='center', valign='top', border=2)
        >>> kwimage.draw_text_on_image(canvas, '(pip install kwimage)', (512, 512 - 10), halign='right', valign='bottom', border=2, fontScale=0.8)
        >>> result = kwimage.nodata_checkerboard(canvas, on_value=0.5)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(result)
        >>> kwplot.show_if_requested()

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> # Simple test with a masked array
        >>> import kwimage
        >>> data = kwimage.grab_test_image(space='rgb', dsize=(64, 64))
        >>> data = kwimage.ensure_uint255(data)
        >>> circle = kwimage.Polygon.circle((32, 32), 16)
        >>> mask = circle.fill(np.zeros(data.shape, dtype=np.uint8), value=1).astype(bool)
        >>> img = np.ma.MaskedArray(data, mask)
        >>> canvas = img.copy()
        >>> result = kwimage.nodata_checkerboard(canvas)
        >>> canvas.data is result.data
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(result, title='nodata_checkers with masked uint8')
        >>> kwplot.show_if_requested()

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> # Simple test with a bad mask
        >>> import kwimage
        >>> data = kwimage.grab_test_image(space='rgb', dsize=(64, 64))
        >>> data = kwimage.ensure_uint255(data)
        >>> circle = kwimage.Polygon.circle((32, 32), 16)
        >>> mask = circle.fill(np.zeros(data.shape, dtype=np.uint8), value=1).astype(bool)
        >>> img = np.ma.MaskedArray(data, mask)
        >>> img.__dict__['_mask'] = np.empty((), dtype=bool)
        >>> import pytest
        >>> with pytest.raises(Exception):
        ...     result = kwimage.nodata_checkerboard(img)
    """
    is_masked = isinstance(canvas, np.ma.MaskedArray)
    masks = []
    if is_masked:
        if canvas.mask.shape != canvas.data.shape:
            # It is unclear if we should error on this case or not
            raise Exception('masked input doesnt agree with data')
        else:
            masks.append(canvas.mask)
        out_canvas = canvas.data
    else:
        out_canvas = canvas

    if out_canvas.dtype.kind == 'f':
        masks.append(np.isnan(out_canvas))

    if masks:
        invalid_mask = np.logical_or.reduce(masks)
    else:
        invalid_mask = None

    if invalid_mask is not None:
        out_canvas = _masked_checkerboard(
            out_canvas, invalid_mask, square_shape, on_value, off_value)

    if is_masked:
        out_canvas = np.ma.MaskedArray(data=out_canvas, mask=invalid_mask)

    return out_canvas



from numpy import ndarray
from _typeshed import Incomplete


def radial_fourier_mask(img_hwc: ndarray,
                        radius: int = ...,
                        axis: Incomplete | None = ...,
                        clip: Incomplete | None = ...):
    ...


def fourier_mask(img_hwc: ndarray,
                 mask: ndarray,
                 axis: Incomplete | None = ...,
                 clip: Incomplete | None = ...,
                 backend: str = ...):
    ...



def test_detections_to_from_coco_missing_parts():
    """
    Test for an issue resolved in 0.11.2, where a keypoint entry was not
    appended to its respective column causing each column to have different
    lengths, and thus causing an issue with zip
    """
    import kwimage
    anns = [
        {
            'id': 0,
            'image_id': 1,
            'category_name': 'cat2',
            'bbox': [2, 3, 10, 10],
        },
        {
            'id': 0,
            'image_id': 1,
            'category_name': 'cat2',
            'bbox': [2, 3, 10, 10],
            'keypoints': [4.5, 4.5, 2],
            'segmentation': kwimage.Polygon.random().to_coco(),
        },
        {
            'id': 0,
            'image_id': 1,
            'category_name': 'cat2',
            'bbox': [2, 3, 10, 10],
            'segmentation': kwimage.Polygon.random().to_coco(),
        },
        {
            'id': 0,
            'image_id': 1,
            'category_name': 'cat2',
            'bbox': [2, 3, 10, 10],
            'keypoints': [4.5, 4.5, 2],
        },
        {
            'id': 0,
            'image_id': 1,
            'category_name': 'cat2',
            'bbox': [2, 3, 10, 10],
            'keypoints': [4.5, 4.5, 2],
        },
    ]
    dets = kwimage.Detections.from_coco_annots(anns)
    # Check missing entries were imputed as None
    assert dets.data['keypoints'].data[0] is None
    assert dets.data['keypoints'].data[1] is not None
    assert dets.data['keypoints'].data[2] is None
    assert dets.data['segmentations'].data[0] is None
    assert dets.data['segmentations'].data[1] is not None
    # Roundtrip conversion
    anns_recon = list(dets.to_coco())
    dets_recon = kwimage.Detections.from_coco_annots(anns_recon)
    anns_recon2 = list(dets_recon.to_coco())
    assert anns_recon == anns_recon2


def test_detections_draw_on_corner_cases():
    import kwimage
    dets = kwimage.Detections.random().scale((64, 64))

    # Baseline, should work correctly.
    canvas = dets.draw_on()

    if 0:
        # for developers
        import kwplot
        kwplot.autompl()
        kwplot.imshow(canvas)

    # Replace class_idxs with null data, and ensure we don't break.
    dets.data['class_idxs'] = [None] * len(dets)
    canvas1 = dets.draw_on()  # NOQA

    # Use cids null data, and ensure we don't break.
    dets.data.pop('class_idxs')
    dets.data['cids'] = [None] * len(dets)
    canvas2 = dets.draw_on()  # NOQA



from typing import Iterable
from numpy import ndarray
from typing import Tuple
from numbers import Number
from typing import Dict
from nptyping import UInt8
from typing import Any
import kwimage
import numpy
from _typeshed import Incomplete


def imscale(img,
            scale,
            interpolation: Incomplete | None = ...,
            return_scale: bool = ...) -> None:
    ...


def imcrop(img: ndarray,
           dsize: Tuple[None | int, None | int],
           about: Tuple[str | int, str | int] | None = None,
           origin: Tuple[int, int] | None = None,
           border_value: Number | Tuple | str | None = None,
           interpolation: str = 'nearest') -> ndarray:
    ...


DTYPE_KEY_TO_DTYPE: Incomplete
DTYPE_TO_DTYPE_KEY: Incomplete
CV2_ALLOWED_DTYPE_MAPPINGS: Incomplete


def imresize(
    img: ndarray,
    scale: float | Tuple[float, float] | None = None,
    dsize: Tuple[int | None, int | None] | None = None,
    max_dim: int | None = None,
    min_dim: int | None = None,
    interpolation: str | int | None = None,
    grow_interpolation: str | int | None = None,
    letterbox: bool = False,
    return_info: bool = False,
    antialias: bool = False,
    border_value: int | float | Iterable[int | float] = 0
) -> ndarray | Tuple[ndarray, Dict]:
    ...


def convert_colorspace(img: ndarray,
                       src_space: str,
                       dst_space: str,
                       copy: bool = ...,
                       implicit: bool = False,
                       dst: ndarray[Any, UInt8] | None = None) -> ndarray:
    ...


def gaussian_patch(
        shape: Tuple[int, int] = ...,
        sigma: float | Tuple[float, float] | None = None) -> ndarray:
    ...


def gaussian_blur(image: ndarray,
                  kernel: int | Tuple[int, int] | None = None,
                  sigma: float | Tuple[float, float] | None = None,
                  border_mode: str | int | None = None,
                  dst: ndarray | None = None) -> ndarray:
    ...


def warp_affine(image: ndarray,
                transform: ndarray | dict | kwimage.Affine,
                dsize: Tuple[int, int] | None | str = None,
                antialias: bool = ...,
                interpolation: str | int = 'linear',
                border_mode: str | int | None = None,
                border_value: int | float | Iterable[int | float] = 0,
                large_warp_dim: int | None | str = None,
                return_info: bool = False) -> ndarray | Tuple[ndarray, Dict]:
    ...


def morphology(data,
               mode: str,
               kernel: ndarray | int | Tuple[int, int] = 5,
               element: str = 'rect',
               iterations: int = 1,
               border_mode: str | int = 'constant',
               border_value: int | float | Iterable[int | float] = 0):
    ...


def connected_components(image: ndarray,
                         connectivity: int = 8,
                         ltype: numpy.dtype | str | int = ...,
                         with_stats: bool = ...,
                         algo: str = 'default') -> Tuple[ndarray, dict]:
    ...


def warp_projective(
        image: ndarray,
        transform: ndarray | dict | kwimage.Projective,
        dsize: Tuple[int, int] | None | str = None,
        antialias: bool = ...,
        interpolation: str | int = 'linear',
        border_mode: str | int | None = None,
        border_value: int | float | Iterable[int | float] = 0,
        large_warp_dim: int | None | str = None,
        return_info: bool = False) -> ndarray | Tuple[ndarray, Dict]:
    ...


def warp_image(image: ndarray,
               transform: ndarray | dict | kwimage.Matrix,
               dsize: Tuple[int, int] | None | str = None,
               antialias: bool = ...,
               interpolation: str | int = 'linear',
               border_mode: str | int | None = None,
               border_value: int | float | Iterable[int | float] = 0,
               large_warp_dim: int | None | str = None,
               return_info: bool = False) -> ndarray | Tuple[ndarray, Dict]:
    ...



from typing import Tuple
from numpy import ndarray
from numbers import Number


def grab_test_image(key: str = 'astro',
                    space: str = 'rgb',
                    dsize: Tuple[int, int] | None = None,
                    interpolation: str = ...) -> ndarray:
    ...


def grab_test_image_fpath(key: str = 'astro',
                          dsize: None | Tuple[int, int] = None,
                          overviews: None | int = None) -> str:
    ...


def checkerboard(num_squares: int | str = 'auto',
                 square_shape: int | Tuple[int, int] | str = 'auto',
                 dsize: Tuple[int, int] = ...,
                 dtype: type = float,
                 on_value: Number | int = 1,
                 off_value: Number | int = 0):
    ...
